
XIncludeFile "StringInclude.pb"
XIncludeFile "Object.pb"
XIncludeFile "Error.pb"
XIncludeFile "Set.pb"
XIncludeFile "CharTable.pb"


Enumeration -1 Step -1
  #GrammarItem_NonTerm ;<=
  #GrammarItem_Start = #GrammarItem_NonTerm
  #GrammarItem_Error
EndEnumeration
#GrammarItem_Term = #Symbol_Epsilon+1 ;>=
#GrammarItem_EndOfFile = #GrammarItem_Term

#GrammarAssoc_Left  = 1
#GrammarAssoc_Right = 2


Structure GrammarItem
  idx.i
  type.s
  priority.i
  assoc.i    ; associativity
  *first.Set
  firstIsEps.i
EndStructure


Structure GrammarTerm Extends Object
  idx.i
EndStructure


Structure GrammarRule Extends Object
  *grammar.Grammar
  runidx.i
  cntvalues.i
  point.i
  *start.GrammarItem
  List *item.GrammarItem()
  *precedence.GrammarItem
  *look.Set
  lookReady.i
EndStructure


Structure Grammar Extends Object
  *rules.Set
  Map nonterm.GrammarItem()
  Map termsym.GrammarItem()
  Array symbol.i(255)
EndStructure


Declare NewGrammarTerm(idx.i)
Declare.s GrammarRuleTermString(*grammar.Grammar, idx.i)
Declare.s GrammarRuleToString(*rule.GrammarRule)

XIncludeFile "LRParser.pb"



;##############################################################################################################
;##### GrammarTerm #####

Procedure GrammarTermCompare(*this.GrammarTerm, *other.GrammarTerm)
  ProcedureReturn *this\idx - *other\idx
EndProcedure


Procedure.s GrammarTermToString(*this.GrammarTerm)
  ProcedureReturn Str(*this\idx)
EndProcedure


Procedure NewGrammarTerm(idx.i)
  Protected *this.GrammarTerm
  
  *this = AllocateMemory(SizeOf(GrammarTerm))
  *this\free = @StandardFreeFunction()
  *this\copy = @StandardCopyFunction()
  *this\compare = @GrammarTermCompare()
  *this\tostring = @GrammarTermToString()
  
  *this\idx = idx
  
  ProcedureReturn *this
EndProcedure


;##############################################################################################################
;##### GrammarRule #####

Procedure GrammarRuleCompare(*this.GrammarRule, *other.GrammarRule)
  Protected cmp
  
  cmp = *this\runidx - *other\runidx
  If (cmp <> 0)
    ProcedureReturn cmp
  EndIf
  
  cmp = *this\point - *other\point
  If (cmp <> 0)
    ProcedureReturn cmp
  EndIf
  
  ProcedureReturn 0
EndProcedure


Declare NewGrammarRule(*start, *look=#Null)

Procedure GrammarRuleCopy(*this.GrammarRule)
  Protected *copy.GrammarRule
  
  *copy = NewGrammarRule(*this\start)
  ForEach *this\item()
    AddElement(*copy\item())
    *copy\item() = *this\item()
  Next
  *copy\grammar    = *this\grammar
  *copy\runidx     = *this\runidx
  *copy\cntvalues  = *this\cntvalues
  *copy\point      = *this\point
  *copy\precedence = *this\precedence
  
  ProcedureReturn *copy
EndProcedure


Procedure GrammarRuleFree(*this.GrammarRule)
  FreeObject(*this\look)
  ClearStructure(*this, GrammarRule)
  FreeMemory(*this)
EndProcedure


Procedure.s GrammarRuleTermString(*grammar.Grammar, idx)
  If (*grammar <> #Null)
    ForEach *grammar\termsym()
      If (*grammar\termsym()\idx = idx)
        ProcedureReturn MapKey(*grammar\termsym());+Str(idx)
      EndIf
    Next
    ForEach *grammar\nonterm()
      If (*grammar\nonterm()\idx = idx)
        ProcedureReturn MapKey(*grammar\nonterm());+Str(idx)
      EndIf
    Next
  EndIf
  ProcedureReturn Str(idx)
EndProcedure

Procedure.s GrammarRuleToString(*this.GrammarRule)
  Protected string.s
  Protected *term.GrammarTerm
  Protected i
  
  string = GrammarRuleTermString(*this\grammar, *this\start\idx)+":"
  ForEach *this\item()
    If (*this\point = ListIndex(*this\item()))
      string + " ."
    EndIf
    string + " "+GrammarRuleTermString(*this\grammar, *this\item()\idx)
  Next
  If (*this\point = -1) And (ListSize(*this\item()) = 0)
    string + " ;{ empty ;}"
  EndIf
  
  If (*this\point = ListSize(*this\item()))
    string + " ."
    
    string + " ["
    For i = 0 To *this\look\size - 1
      *term = *this\look\object(i)
      If (i > 0)
        string + ", "
      EndIf
      string + GrammarRuleTermString(*this\grammar, *term\idx)
    Next
    string + "]"
  EndIf
  
  ProcedureReturn string
EndProcedure


Procedure NewGrammarRule(*start, *look=#Null)
  Protected *this.GrammarRule
  
  *this = AllocateMemory(SizeOf(GrammarRule))
  InitializeStructure(*this, GrammarRule)
  *this\free = @GrammarRuleFree()
  *this\copy = @GrammarRuleCopy()
  *this\compare = @GrammarRuleCompare()
  *this\tostring = @GrammarRuleToString()
  
  *this\start = *start
  If (*look = #Null)
    *this\look = NewSet()
  Else
    *this\look = *look
  EndIf
  
  ProcedureReturn *this
EndProcedure


;##############################################################################################################
;##### Grammar #####

Procedure GrammarCompare(*this.Grammar, *other.Grammar)
  ProcedureReturn *this\rules\compare(*this\rules, *other\rules)
EndProcedure


Declare NewGrammar(*rules.Set=#Null)

Procedure GrammarCopy(*this.Grammar)
  Protected *copy.Grammar
  
  *copy = NewGrammar(CopyObject(*this\rules))
  
  ProcedureReturn *copy
EndProcedure


Procedure GrammarFree(*this.Grammar)
  FreeObject(*this\rules)
  ClearStructure(*this, Grammar)
  FreeMemory(*this)
EndProcedure


Procedure.s GrammarToString(*this.Grammar)
  Protected string.s
  Protected i
  string = "{"
  
  For i = 0 To *this\rules\size - 1
    If (i > 0)
      string + ", "
    EndIf
    string + Chr(10)
    string + ToStringObject(*this\rules\object(i))
  Next
  
  string + Chr(10)
  string + "}"
  ProcedureReturn string
EndProcedure


Procedure GrammarAddRule(*this.Grammar, *rule.GrammarRule, *isadd.Byte=#Null)
  Protected *result, isadd.Byte
  
  If (*isadd = #Null)
    *isadd = isadd
  EndIf
  
  If (ListSize(*rule\item()) > 0)
    LastElement(*rule\item())
    If (*rule\item() = #Null)
      DeleteElement(*rule\item())
    EndIf
  EndIf
  
  *rule\grammar = *this
  *rule\cntvalues = ListSize(*rule\item())
  *result = SetAddObject(*this\rules, *rule, *isadd)
  If (*isadd\b = #False)
    FreeObject(*rule)
  EndIf
  
  ProcedureReturn *rule
EndProcedure


Procedure GrammarAddTermSym(*grammar.Grammar, text.s, type.s)
  AddMapElement(*grammar\termsym(), text)
  *grammar\termsym()\idx = #GrammarItem_Term + MapSize(*grammar\termsym()) - 1
  *grammar\termsym()\type = type
EndProcedure


Procedure GrammarAddNonTerm(*grammar.Grammar, text.s, type.s)
  AddMapElement(*grammar\nonterm(), text)
  *grammar\nonterm()\idx = #GrammarItem_NonTerm - MapSize(*grammar\nonterm()) + 1
  *grammar\nonterm()\type = type
EndProcedure


Procedure GrammarStartEval(*this.Grammar, string.s)
  Protected *char.Character, *start
  Protected nonterm.s, *rule.GrammarRule
  
  *char = @string
  
  While (IsIgnoreChar(*char\c))
    *char + SizeOf(Character)
  Wend
  *start = *char
  
  If (*char\c >= 'a') And (*char\c <= 'z')
    *char + SizeOf(Character)
    While (IsNameChar(*char\c) Or IsNumberChar(*char\c))
      *char + SizeOf(Character)
    Wend
    
    nonterm = PeekS(*start, (*char-*start) / SizeOf(Character))
    While (IsIgnoreChar(*char\c))
      *char + SizeOf(Character)
    Wend
    
    If (*char\c = ':')
      If (FindMapElement(*this\nonterm(), nonterm) = 0)
        GrammarAddNonTerm(*this, nonterm, "")
      EndIf
      *rule = NewGrammarRule(*this\nonterm())
      *rule\grammar = *this
      ProcedureReturn *rule
    EndIf
  EndIf
EndProcedure


Procedure GrammarToLRParser(*grammar.Grammar)
  Protected *parser.NFA
  Protected *rule.GrammarRule
  Protected *set.Set
  Protected i
  
  *parser = NewNFA(#Null)
  
  *rule = *grammar\rules\object(0)
  SetAddObject(*rule\look, NewGrammarTerm(#GrammarItem_EndOfFile))
  
  *set = NewSet()
  SetAddObject(*set, *rule)
  LRParserClosure(*grammar, *set)
  LRParserAddState(*grammar, *parser, *set)
  LRParserLookahead(*grammar, *parser)
  
  ProcedureReturn *parser
EndProcedure


Procedure NewGrammar(*rules.Set=#Null)
  Protected *this.Grammar
  
  *this = AllocateMemory(SizeOf(Grammar))
  InitializeStructure(*this, Grammar)
  *this\free = @GrammarFree()
  *this\copy = @GrammarCopy()
  *this\compare = @GrammarCompare()
  *this\tostring = @GrammarToString()
  
  If (*rules = #Null)
    *this\rules = NewSet()
  Else
    *this\rules = *rules
  EndIf
  
  ProcedureReturn *this
EndProcedure


; IDE Options = PureBasic 4.61 (Windows - x86)
; EnableXP