
XIncludeFile "Object.pb"
XIncludeFile "Set.pb"
XIncludeFile "NFA.pb"


#LRParser_Accept = 1
#LRParser_Shift  = $10000000
#LRParser_Reduce = $20000000


Structure LRParserSyntaxTable
  idx.i
  StructureUnion
    *trans.NFATrans
    *rule
  EndStructureUnion
EndStructure


Structure LRParserConflict
  StructureUnion
    *trans.NFATrans
    *ruleA.GrammarRule
  EndStructureUnion
  *ruleB.GrammarRule
  term.i
EndStructure


Structure LRParserState
  Array action.LRParserSyntaxTable(0)
  Array gotot.LRParserSyntaxTable(0)
  List rrconflict.LRParserConflict()
  List srconflict.LRParserConflict()
EndStructure


Global Dim parserState.LRParserState(0)


Declare LRParserToDot(*parser.NFA, *grammar.Grammar, file.s)
Declare LRParserRuleFirst(*grammar.Grammar, *rule.GrammarRule, point.i, *look.Set, *isEps.Byte=#Null)


;##############################################################################################################
;##### LRParser #####


Structure LRParserLookQueue
  *state.NFAState
  *rule.GrammarRule
EndStructure


Procedure LRParserSymFirst(*grammar.Grammar, idx, *isEps.Byte)
  Protected *nonterm.GrammarItem
  Protected *rule.GrammarRule
  Protected *first.Set, *newset.Set
  Protected nisEps.Byte
  Protected i, k
  
  *isEps\b = #False
  
  If (idx >= #GrammarItem_Term)
    ForEach *grammar\termsym()
      If (*grammar\termsym()\idx = idx)
        If (*grammar\termsym()\first = #Null)
          *grammar\termsym()\first = NewSet()
          SetAddObject(*grammar\termsym()\first, NewGrammarTerm(idx))
        EndIf
        ProcedureReturn *grammar\termsym()\first
      EndIf
    Next
    
  Else
    ForEach *grammar\nonterm()
      If (*grammar\nonterm()\idx = idx)
        *nonterm = *grammar\nonterm()
        *first = *nonterm\first
        
        If (*first = #Null)
          *first = NewSet()
          *nonterm\first = *first
          
          For i = 0 To *grammar\rules\size - 1
            *rule = *grammar\rules\object(i)
            If (*rule\start\idx = idx)
              *newset = LRParserRuleFirst(*grammar, *rule, 0, #Null, nisEps)
              
              For k = 0 To *newset\size - 1
                SetAddObject(*first, *newset\object(k))
              Next
              If (nisEps\b = #True)
                *nonterm\firstIsEps = #True
              EndIf
            EndIf
          Next
        EndIf
        
        *isEps\b = *nonterm\firstIsEps
        ProcedureReturn *nonterm\first
      EndIf
    Next
    
  EndIf
EndProcedure


Procedure LRParserRuleFirst(*grammar.Grammar, *rule.GrammarRule, point.i, *look.Set, *isEps.Byte=#Null)
  Protected *set.Set, *newset.Set
  Protected nisEps.Byte
  Protected i
  
  If (*isEps <> #Null)
    *isEps\b = #False
  EndIf
  *set = NewSet()
  
  If (point = 0)
    ResetList(*rule\item())
  Else
    SelectElement(*rule\item(), point-1)
  EndIf
  While NextElement(*rule\item())
    *newset = LRParserSymFirst(*grammar, *rule\item()\idx, nisEps)
    
    For i = 0 To *newset\size - 1
      SetAddObject(*set, *newset\object(i))
    Next
    
    If (nisEps\b = #False)
      ProcedureReturn *set
    EndIf
  Wend
  
  If (*look = #Null)
    *isEps\b = #True
  Else
    For i = 0 To *look\size - 1
      SetAddObject(*set, *look\object(i))
    Next
  EndIf
  
  ProcedureReturn *set
EndProcedure


Procedure LRParserLookahead(*grammar.Grammar, *parser.NFA)
  Protected *state.NFAState, *trans.NFATrans
  Protected *set.Set, *gset.Set, *first.Set
  Protected *rule.GrammarRule, *grule.GrammarRule
  Protected isadd.Byte
  Protected i, k, m
  
  Protected NewList lookQueue.LRParserLookQueue()
  
  For i = 0 To *parser\state\size - 1
    *state = *parser\state\object(i)
    *set = *state\value
    
    For k = 0 To *set\size - 1
      *rule = *set\object(k)
      
      AddElement(lookQueue())
      lookQueue()\rule  = *rule
      lookQueue()\state = *state
      *rule\lookReady = #False
    Next
  Next
  
  While (ListSize(lookQueue()) > 0)
    FirstElement(lookQueue())
    *state = lookQueue()\state
    *rule  = lookQueue()\rule
    *rule\lookReady = #True
    DeleteElement(lookQueue())
    
    If (*rule\point < ListSize(*rule\item()))
      *first = #Null
      *set = *state\value
      
      For k = 0 To *set\size - 1
        *grule = *set\object(k)
        
        SelectElement(*rule\item(), *rule\point)
        If (*rule\item()\idx = *grule\start\idx)
          If (*first = #Null)
            *first = LRParserRuleFirst(*grammar, *rule, *rule\point+1, *rule\look)
          EndIf
          
          For m = 0 To *first\size - 1
            SetAddObject(*grule\look, *first\object(m), isadd)
            
            If (isadd\b = #True)
              If (*grule\lookReady = #True)
                LastElement(lookQueue())
                AddElement(lookQueue())
                lookQueue()\rule  = *grule
                lookQueue()\state = *state
                *grule\lookReady = #False
              EndIf
            EndIf
          Next
        EndIf
      Next
    EndIf
    
    For i = 0 To *state\trans\size - 1
      *trans = *state\trans\object(i)
      *gset = *trans\goal\value
      
      For k = 0 To *gset\size - 1
        *grule = *gset\object(k)
        
        If (*rule\runidx = *grule\runidx) And (*rule\point = *grule\point - 1)
          For m = 0 To *rule\look\size - 1
            SetAddObject(*grule\look, *rule\look\object(m), isadd)
            
            If (isadd\b = #True)
              If (*grule\lookReady = #True)
                LastElement(lookQueue())
                AddElement(lookQueue())
                lookQueue()\rule  = *grule
                lookQueue()\state = *trans\goal
                *grule\lookReady = #False
              EndIf
            EndIf
          Next
          
          Break
        EndIf
      Next
    Next
  Wend
EndProcedure


Global NewMap *ClosureHistory.Set()

Procedure LRParserClosure(*grammar.Grammar, *set.Set)
  Protected.GrammarRule *rule, *gramrule, *gramnewr, *gramaddr
  Protected isadd.Byte, iseps.Byte
  Protected again, idx, point
  Protected i, k, m
  Protected key.s
  
  key = Str(*set\size)+";"
  For i = 0 To *set\size - 1
    *rule = *set\object(i)
    key + Str(*rule\runidx)+","
    key + Str(*rule\point)+";"
  Next
  If FindMapElement(*ClosureHistory(), key)
    ProcedureReturn *ClosureHistory()
  EndIf
  
  Repeat
    again = #False
    For i = 0 To *set\size - 1
      *rule = *set\object(i)
      
      If (*rule\point < ListSize(*rule\item()))
        SelectElement(*rule\item(), *rule\point)
        idx = *rule\item()\idx
        
        If (idx <= #GrammarItem_NonTerm)
          For k = 0 To *grammar\rules\size - 1
            *gramrule = *grammar\rules\object(k)
            
            If (*gramrule\start\idx = idx)
              point = *gramrule\point
              *gramrule\point = 0
              *gramaddr = SetAddObject(*set, *gramrule, isadd)
              
              If (isadd\b = #True)
                *gramnewr = CopyObject(*gramrule)
                For m = 0 To *set\size - 1
                  If (*set\object(m) = *gramaddr)
                    *set\object(m) = *gramnewr
                  EndIf
                Next
                again = #True
              EndIf
              
              *gramrule\point = point
            EndIf
          Next
        EndIf
      EndIf
    Next
  Until (again = #False)
;   
  AddMapElement(*ClosureHistory(), key)
  *ClosureHistory() = *set
  
  ProcedureReturn *set
EndProcedure


Procedure LRParserGoto(*grammar.Grammar, *rules.Set, symbol.i)
  Protected NewList *rules.GrammarRule()
  Protected *rule.GrammarRule, *newr.GrammarRule, *set.Set
  Protected i
  
  *set = NewSet()
  
  For i = 0 To *rules\size - 1
    *rule = *rules\object(i)
    
    If (*rule\point < ListSize(*rule\item()))
      SelectElement(*rule\item(), *rule\point)
      If (*rule\item()\idx = symbol)
        *newr = CopyObject(*rule)
        *newr\point + 1
        SetAddObject(*set, *newr)
      EndIf
    EndIf
  Next
  
  *set = LRParserClosure(*grammar, *set)
  ProcedureReturn *set
EndProcedure


Procedure LRParserAddState(*grammar.Grammar, *parser.NFA, *set.Set, *transStart.NFAState=#Null, transValue.i=0)
  Protected NewList *item.GrammarItem()
  Protected *rule.GrammarRule, ruleidx, isalready, *newset
  Protected *state.NFAState, *save, isadd.Byte
  Protected *trans.NFATrans
  Protected i
  
  *state = NFAAddState(*parser, *set, isadd)
  
  If (*transStart <> #Null)
    NFAStateAddTrans(*transStart, *state, transValue)
  EndIf
  
  If (isadd\b = #True)
    For i = 0 To *set\size - 1
      *rule = *set\object(i)
      
      If (*rule\point < ListSize(*rule\item()))
        SelectElement(*rule\item(), *rule\point)
        ruleidx = *rule\item()\idx
        
        isalready = #False
        ForEach *item()
          If (*item() = *rule\item())
            isalready = #True
            Break
          EndIf
        Next
        
        If (isalready = #False)
          AddElement(*item())
          *item() = *rule\item()
          *newset = LRParserGoto(*grammar, *set, ruleidx)
          LRParserAddState(*grammar, *parser, *newset, *state, ruleidx)
        EndIf
      EndIf
    Next
  EndIf
  
  ProcedureReturn *state
EndProcedure


Procedure LRParserToSyntaxTable(*parser.NFA, *grammar.Grammar)
  Protected cntstates, cnttermsym, cntnonterm, s, k
  Protected *state.NFAState, *trans.NFATrans
  Protected *rules.Set, *rule.GrammarRule, *term.GrammarTerm
  Protected *itemA.GrammarItem, *itemB.GrammarItem
  Protected i, m, n
  
  cntstates  = *parser\state\size
  cnttermsym = MapSize(*grammar\termsym())
  cntnonterm = MapSize(*grammar\nonterm())
  
  Dim parserState(cntstates-1)
  For s = 0 To cntstates - 1
    Dim parserState(s)\action(cnttermsym-1)
    Dim parserState(s)\gotot (cntnonterm-1)
  Next

  For i = 0 To *parser\state\size - 1
    *state = *parser\state\object(i)
    s = *state\number
    
    For m = 0 To *state\trans\size - 1
      *trans = *state\trans\object(m)
      
      If (*trans\value >= #GrammarItem_Term)
        k = *trans\value - #GrammarItem_Term
        parserState(s)\action(k)\idx = #LRParser_Shift | *trans\goal\number
        parserState(s)\action(k)\trans = *trans
        
      ElseIf (*trans\value <= #GrammarItem_NonTerm)
        k = #GrammarItem_NonTerm - *trans\value
        parserState(s)\gotot(k)\idx = #LRParser_Shift | *trans\goal\number
        parserState(s)\gotot(k)\trans = *trans
        
      EndIf
    Next
    
    *rules = *state\value
    For m = 0 To *rules\size - 1
      *rule = *rules\object(m)
      
      If (*rule\point = ListSize(*rule\item()))
        For n = 0 To *rule\look\size - 1
          *term = *rule\look\object(n)
          
          If (*term\idx >= #GrammarItem_Term)
            k = *term\idx - #GrammarItem_Term
          Else
            Continue
          EndIf
          
          If (parserState(s)\action(k)\idx & #LRParser_Shift)
            
            *itemA = #Null
            *itemB = #Null
            ForEach *grammar\termsym()
              If (*grammar\termsym()\idx = parserState(s)\action(k)\trans\value)
                *itemB = *grammar\termsym()
                Break
              EndIf
            Next
            If (*itemB <> #Null) And (*itemB\priority > 0)
              If (LastElement(*rule\item()) > 0)
                If (*rule\precedence <> #Null)
                  *itemA = *rule\precedence
                Else
                  Repeat
                    If (*rule\item()\idx >= #GrammarItem_Term)
                      Break
                    EndIf
                  Until (PreviousElement(*rule\item()) = #Null)
                  If (*rule\item()\priority > 0)
                    *itemA = *rule\item()
                  EndIf
                EndIf
              EndIf
            EndIf
            If (*itemA <> #Null)
              If (*itemA\priority > *itemB\priority)
                parserState(s)\action(k)\idx = #LRParser_Reduce | *rule\runidx
                parserState(s)\action(k)\rule = *rule
              ElseIf (*itemA\priority = *itemB\priority)
                If (*itemA\assoc = #GrammarAssoc_Left)
                  parserState(s)\action(k)\idx = #LRParser_Reduce | *rule\runidx
                  parserState(s)\action(k)\rule = *rule
                EndIf
              EndIf
            Else
              AddElement(parserState(s)\srconflict())
              parserState(s)\srconflict()\trans = parserState(s)\action(k)\trans
              parserState(s)\srconflict()\ruleB = *rule
              parserState(s)\srconflict()\term  = k
              ;actionTable(s, k)\idx = #LRParser_Reduce | *rule\runidx
            EndIf
            
          ElseIf (parserState(s)\action(k)\idx & #LRParser_Reduce)
            AddElement(parserState(s)\rrconflict())
            parserState(s)\rrconflict()\ruleA = parserState(s)\action(k)\rule
            parserState(s)\rrconflict()\ruleB = *rule
            parserState(s)\rrconflict()\term  = k
            
            If (parserState(s)\action(k)\idx > (#LRParser_Reduce | *rule\runidx))
              Swap parserState(s)\rrconflict()\ruleA, parserState(s)\rrconflict()\ruleB
              parserState(s)\action(k)\idx = #LRParser_Reduce | *rule\runidx
              parserState(s)\action(k)\rule = *rule
            EndIf
            
          Else
            If (*rule\runidx = 0)
              parserState(s)\action(k)\idx = #LRParser_Accept
            Else
              parserState(s)\action(k)\idx = #LRParser_Reduce | *rule\runidx
              parserState(s)\action(k)\rule = *rule
            EndIf
          EndIf
            
        Next
      EndIf
    Next
  Next
EndProcedure


Procedure LRParserGetStart(*parser.NFA, *grammar.Grammar)
  Protected *state.NFAState, *rules.Set, *rule.GrammarRule
  Protected i, k
  
  For i = 0 To *parser\state\size - 1
    *state = *parser\state\object(i)
    *rules = *state\value
    
    For k = 0 To *rules\size - 1
      *rule = *rules\object(k)
      
      If (*rule\start\idx = #GrammarItem_Start) And (*rule\point = 0)
        ProcedureReturn *state\number
      EndIf
    Next
  Next
EndProcedure


Structure LRParserDebugSort
  term.i
  str.s
  idx.i
  conflict.i
EndStructure

Procedure LRParserDebugOutput(*parser.NFA, *grammar.Grammar, out)
  Protected NewList sort.LRParserDebugSort()
  Protected cnttermsym, cntnonterm, maxspace
  Protected *state.NFAState, *rules.Set, *rule.GrammarRule
  Protected *trans.NFATrans
  Protected i, k, r, s, srconf, rrconf, term
  Protected text.s, point, maxrule, laststart, pos
  
  Protected Dim rulecnt.i(*grammar\rules\size - 1)
  
  cnttermsym = MapSize(*grammar\termsym())
  cntnonterm = MapSize(*grammar\nonterm())
  
  For s = 0 To *parser\state\size - 1
    srconf = ListSize(parserState(s)\srconflict())
    rrconf = ListSize(parserState(s)\rrconflict())
    
    If (srconf <> 0) Or (rrconf <> 0)
      text = "State "+Str(s)+" conflicts: "
      If (srconf <> 0)
        text + Str(srconf) + " shift/reduce"
      EndIf
      If (srconf <> 0) And (rrconf <> 0)
        text + ", "
      EndIf
      If (rrconf <> 0)
        text + Str(rrconf) + " reduce/reduce"
      EndIf
      WriteStringN(out, text)
    EndIf
  Next
  WriteStringN(out, "")
  WriteStringN(out, "")
  
  WriteStringN(out, "Grammar: ")
  laststart = #PB_Ignore
  For r = 0 To *grammar\rules\size - 1
    *rule = *grammar\rules\object(r)
    
    point = *rule\point
    *rule\point = -1
    text = GrammarRuleToString(*rule)
    If (*rule\start\idx = laststart)
      pos = FindString(text,":")
      text = Space(pos-1) + "|" + Mid(text,pos+1)
    EndIf
    If (laststart <> #PB_Ignore) And (laststart <> *rule\start\idx)
      WriteStringN(out, "")
    EndIf
    WriteStringN(out, RSet(Str(r),5," ") + " " + text)
    laststart = *rule\start\idx
    *rule\point = point
  Next
  WriteStringN(out, "")
  WriteStringN(out, "")
  
  For s = 0 To *parser\state\size - 1
    For i = 0 To *parser\state\size - 1
      *state = *parser\state\object(i)
      If (s = *state\number)
        
        WriteStringN(out, "State "+Str(s))
        WriteStringN(out, "")
        
        *rules = *state\value
        
        ; rules
        laststart = #PB_Ignore
        For k = 0 To *rules\size - 1
          *rule = *rules\object(k)
          r = *rule\runidx
          
          text = GrammarRuleToString(*rule)
          If (*rule\start\idx = laststart)
            pos = FindString(text,":")
            text = Space(pos-1) + "|" + Mid(text,pos+1)
          EndIf
          WriteStringN(out, RSet(Str(r),5," ") + " " + text)
          laststart = *rule\start\idx
        Next
        If (*grammar\rules\size > 0)
          WriteStringN(out, "")
        EndIf
        
        ; shift with terminal
        ClearList(sort())
        maxspace = 0
        For k = 0 To cnttermsym - 1
          If (parserState(s)\action(k)\idx & #LRParser_Shift)
            *trans = parserState(s)\action(k)\trans
            AddElement(sort())
            sort()\str = GrammarRuleTermString(*grammar, *trans\value)
            sort()\idx = parserState(s)\action(k)\idx - #LRParser_Shift
            If (Len(sort()\str) > maxspace)
              maxspace = Len(sort()\str)
            EndIf
          EndIf
        Next
        SortStructuredList(sort(), #PB_Sort_Ascending, OffsetOf(LRParserDebugSort\idx), #PB_Sort_Integer)
        ForEach sort()
          WriteStringN(out, "    "+LSet(sort()\str,maxspace," ")+"  shift and goto state "+Str(sort()\idx))
        Next
        If (maxspace > 0)
          WriteStringN(out, "")
        EndIf
        
        ; reduce
        Dim rulecnt.i(*grammar\rules\size - 1)
        ClearList(sort())
        maxspace = 0
        For k = 0 To cnttermsym - 1
          If (parserState(s)\action(k)\idx & #LRParser_Reduce)
            *rule = parserState(s)\action(k)\rule
            AddElement(sort())
            sort()\term = k
            sort()\str  = GrammarRuleTermString(*grammar, *rule\start\idx)
            sort()\idx  = *rule\runidx
            rulecnt(*rule\runidx) + 1
          EndIf
        Next
        ForEach parserState(s)\srconflict()
          *rule = parserState(s)\srconflict()\ruleB
          AddElement(sort())
          sort()\term = parserState(s)\srconflict()\term
          sort()\str  = GrammarRuleTermString(*grammar, *rule\start\idx)
          sort()\idx  = *rule\runidx
          sort()\conflict = #True
        Next
        ForEach parserState(s)\rrconflict()
          *rule = parserState(s)\rrconflict()\ruleB
          AddElement(sort())
          sort()\term = parserState(s)\rrconflict()\term
          sort()\str  = GrammarRuleTermString(*grammar, *rule\start\idx)
          sort()\idx  = *rule\runidx
          sort()\conflict = #True
          rulecnt(*rule\runidx) - 1
        Next
        For k = 0 To *grammar\rules\size - 1
          If (maxrule = 0) Or (rulecnt(k) > rulecnt(maxrule))
            maxrule = k
          EndIf
        Next
        If (rulecnt(maxrule) > 1)
          ForEach sort()
            term = sort()\term
            If (sort()\conflict = #False) And (sort()\idx = maxrule)
              PushListPosition(sort())
              While NextElement(sort())
                If (sort()\term = term) And (sort()\conflict = #True)
                  term = -1
                  Break
                EndIf
              Wend
              PopListPosition(sort())
              If (term > -1)
                text = sort()\str
                DeleteElement(sort())
              EndIf
            EndIf
          Next
          AddElement(sort())
          sort()\str  = text
          sort()\term = -1
          sort()\idx  = maxrule
        EndIf
        ForEach sort()
          If (sort()\term = -1)
            text = "$default"
          Else
            text = GrammarRuleTermString(*grammar, #GrammarItem_Term + sort()\term)
          EndIf
          If (Len(text) > maxspace)
            maxspace = Len(text)
          EndIf
        Next
        
        SortStructuredList(sort(), #PB_Sort_Ascending, OffsetOf(LRParserDebugSort\term), #PB_Sort_Integer)
        ForEach sort()
          If (sort()\term = -1)
            text = "$default"
          Else
            text = GrammarRuleTermString(*grammar, #GrammarItem_Term + sort()\term)
          EndIf
          text = "    "+LSet(text, maxspace, " ")+"  "
          If (sort()\conflict)
            text + "["
          EndIf
          text + "reduce with rule "+Str(sort()\idx)+" ("+sort()\str+")"
          If (sort()\conflict)
            text + "]"
          EndIf
          WriteStringN(out, text)
        Next
        If (maxspace > 0)
          WriteStringN(out, "")
        EndIf
        
        ; shift with nonterminal
        ClearList(sort())
        maxspace = 0
        For k = 0 To cntnonterm - 1
          If (parserState(s)\gotot(k)\idx & #LRParser_Shift)
            *trans = parserState(s)\gotot(k)\trans
            AddElement(sort())
            sort()\str = GrammarRuleTermString(*grammar, *trans\value)
            sort()\idx = parserState(s)\gotot(k)\idx - #LRParser_Shift
            If (Len(sort()\str) > maxspace)
              maxspace = Len(sort()\str)
            EndIf
          EndIf
        Next
        SortStructuredList(sort(), #PB_Sort_Ascending, OffsetOf(LRParserDebugSort\idx), #PB_Sort_Integer)
        ForEach sort()
          WriteStringN(out, "    "+LSet(sort()\str,maxspace," ")+"  goto state "+Str(sort()\idx))
        Next
        If (maxspace > 0)
          WriteStringN(out, "")
        EndIf
        
        WriteStringN(out, "")
      EndIf
    Next
  Next
EndProcedure


Procedure LRParserToDot(*parser.NFA, *grammar.Grammar, file.s)
  Protected out, newline, comma, char.c
  Protected *state.NFAState, *trans.NFATrans
  Protected *set.Set, *rule.GrammarRule
  Protected *term.GrammarTerm
  Protected i, k, m
  
  out = CreateFile(#PB_Any, file)
  If (out <> 0)
    WriteStringFormat(out, #PB_UTF8)
    WriteStringN(out, "digraph G {")
    
    For i = 0 To *parser\state\size - 1
      *state = *parser\state\object(i)
      
      WriteStringN(out, "")
      WriteString(out, "  S"+Str(*state\number)+" [shape=record, label="+Chr(34)+"{State "+Str(*state\number))
      If (*parser\start = *state)
        WriteString(out, " - start")
      EndIf
      WriteString(out, " | ")
      
      *set = *state\value
      newline = #False
      For k = 0 To *set\size - 1
        If (newline)
          WriteString(out, "\n")
        EndIf
        *rule = *set\object(k)
        WriteString(out, GrammarRuleTermString(*grammar, *rule\start\idx)+" ")
        WriteString(out, "-\>")
;         char = $2192
;         WriteString(out, Chr(char))
        
        ForEach *rule\item()
          If (ListIndex(*rule\item()) = *rule\point)
            WriteString(out, " .")
          EndIf
          WriteString(out, " "+GrammarRuleTermString(*grammar, *rule\item()\idx))
        Next
        If (ListSize(*rule\item()) = *rule\point)
          WriteString(out, " .")
        EndIf
        
        WriteString(out, " \{")
        comma = #False
        For m = 0 To *rule\look\size - 1
          If (comma)
            WriteString(out, ", ")
          EndIf
          *term = *rule\look\object(m)
          WriteString(out, GrammarRuleTermString(*grammar, *term\idx))
          comma = #True
        Next
        WriteString(out, "\}")
        
        newline = #True
      Next
      
      WriteStringN(out, "}"+Chr(34)+"]")
      
      For k = 0 To *state\trans\size - 1
        *trans = *state\trans\object(k)
        
        WriteString(out, "  S"+Str(*state\number)+" -> S"+Str(*trans\goal\number))
        WriteStringN(out, " [label="+Chr(34)+GrammarRuleTermString(*grammar, *trans\value)+Chr(34)+"]")
      Next
    Next
    
    WriteStringN(out, "}")
    CloseFile(out)
  EndIf
EndProcedure


Macro LRParserWTRuleToString(_rule_)
  savepoint = _rule_\point
  _rule_\point = -1
  warnstr + ToStringObject(_rule_)+Chr(10)
  _rule_\point = savepoint
EndMacro

Procedure.s LRParserWarningText(*parser.NFA, *grammar.Grammar)
  Protected warnstr.s
  Protected s, i, srconf, rrconf
  Protected savepoint
  
  For s = 0 To *parser\state\size - 1
    srconf + ListSize(parserState(s)\srconflict())
    rrconf + ListSize(parserState(s)\rrconflict())
  Next
  
  If (srconf > 0) Or (rrconf > 0)
    warnstr = Str(srconf)+" shift/reduce-conflicts"
    warnstr + Chr(10)
    warnstr + Str(rrconf)+" reduce/reduce-conflicts"
    warnstr + Chr(10)+Chr(10)
    i = 0
    For s = 0 To *parser\state\size - 1
      ForEach parserState(s)\srconflict()
        i + 1
        warnstr + Str(i)+".  State "+Str(s)+Chr(10)
        warnstr + "    Term: "+GrammarRuleTermString(*grammar, #GrammarItem_Term + parserState(s)\srconflict()\term)+Chr(10)
        warnstr + "    "
        LRParserWTRuleToString(parserState(s)\srconflict()\ruleB)
      Next
    Next
    
    warnstr + Chr(10)
    i = 0
    For s = 0 To *parser\state\size - 1
      ForEach parserState(s)\rrconflict()
        i + 1
        warnstr + Str(i)+".  State "+Str(s)+Chr(10)
        warnstr + "    "
        LRParserWTRuleToString(parserState(s)\rrconflict()\ruleA)
        warnstr + "    "
        LRParserWTRuleToString(parserState(s)\rrconflict()\ruleB)
      Next
    Next
    ProcedureReturn warnstr
  EndIf
EndProcedure


; IDE Options = PureBasic 5.00 (Windows - x64)
; EnableXP