
XIncludeFile "CharTable.pb"
XIncludeFile "Set.pb"

;##############################################################################################################

Structure NFATrans Extends Object
  value.i
  *goal.NFAState
EndStructure


Structure NFAState Extends Object
  *value.Object
  number.i
  final.i
  *trans.Set
EndStructure


Structure NFA Extends Object
  *start.NFAState
  *state.Set
  *charTable.CharTable
EndStructure


;##############################################################################################################
;##### NFATrans #####

Procedure NFATransFree(*this.NFATrans)
  FreeMemory(*this)
EndProcedure


Declare NewNFATrans(*goal.NFAState, value.i)

Procedure NFATransCopy(*this.NFATrans)
  ProcedureReturn NewNFATrans(CopyObject(*this\goal), *this\value)
EndProcedure


Procedure NFATransCompare(*this.NFATrans, *other.NFATrans)
  Protected cmp
  
  cmp = *this\goal\compare(*this\goal, *other\goal)
  If (cmp <> 0)
    ProcedureReturn cmp
  EndIf
  
  ProcedureReturn *this\value - *other\value
EndProcedure


Procedure NFATransCompareSymbol(*this.NFATrans, *other.NFATrans)
  ProcedureReturn *this\value - *other\value
EndProcedure


Procedure.s NFATransToString(*this.NFATrans)
  Protected string.s
  
  string = Chr(10)+"->"
  string + ToStringObject(*this\goal)
  string + "("+Str(*this\value)+")"
  
  ProcedureReturn string
EndProcedure


Procedure NewNFATrans(*goal.NFAState, value.i)
  Protected *this.NFATrans
  
  *this = AllocateMemory(SizeOf(NFATrans))
  *this\free = @NFATransFree()
  *this\copy = @NFATransCopy()
  *this\compare = @NFATransCompare()
  *this\tostring = @NFATransToString()
  
  *this\goal  = *goal
  *this\value = value
  
  ProcedureReturn *this
EndProcedure


;##############################################################################################################
;##### NFAState #####

Procedure NFAStateFree(*this.NFAState)
  FreeObject(*this\value)
  FreeObject(*this\trans)
  FreeMemory(*this)
EndProcedure


Declare NewNFAState(*value.Object)

Procedure NFAStateCopy(*this.NFAState)
  Protected *copy.NFAState
  
  *copy = NewNFAState(CopyObject(*this\value))
  *copy\number = *this\number
  *copy\final  = *this\final
  ProcedureReturn *copy
EndProcedure


Procedure NFAStateCompare(*this.NFAState, *other.NFAState)
  ProcedureReturn *this\value\compare(*this\value, *other\value)
EndProcedure


Procedure.s NFAStateToString(*this.NFAState)
  ProcedureReturn ToStringObject(*this\value)
EndProcedure


Procedure NFAStateAddTrans(*this.NFAState, *goal.NFAState, value.i, *isadd.Byte=#Null)
  Protected *trans.NFATrans, *result, isadd.Byte
  
  If (*isadd = #Null)
    *isadd = isadd
  EndIf
  *trans = NewNFATrans(*goal, value)
  
  *result = SetAddObject(*this\trans, *trans, *isadd)
  If (*isadd\b = #False)
    FreeMemory(*trans)
  EndIf
  
  ProcedureReturn *result
EndProcedure


Procedure NewNFAState(*value.Object)
  Protected *this.NFAState
  
  *this = AllocateMemory(SizeOf(NFAState))
  *this\free = @NFAStateFree()
  *this\copy = @NFAStateCopy()
  *this\compare = @NFAStateCompare()
  *this\tostring = @NFAStateToString()
  
  *this\value = *value
  *this\trans = NewSet()
  *this\final = -1
  
  ProcedureReturn *this
EndProcedure


;##############################################################################################################
;##### NFA #####

Declare NewNFA(*charTable.CharTable)

Procedure NFAFree(*this.NFA)
  FreeObject(*this\state)
  If (*this\charTable <> #Null)
    FreeObject(*this\charTable)
  EndIf
  FreeMemory(*this)
EndProcedure

Procedure NFACompare(*this.NFA, *other.NFA)
  ProcedureReturn *this\state\compare(*this\state, *other\state)
EndProcedure


Procedure.s NFAToString(*this.NFA)
  Protected *state.NFAState
  Protected string.s, i
  
  For i = 0 To *this\state\size - 1
    If (i > 0)
      string + Chr(10)
    EndIf
    
    *state = *this\state\object(i)
    string + ToStringObject(*this\state\object(i))
    string + "(final="+Str(*state\final)
    If (*state = *this\start)
      string + ":start"
    EndIf
    
    string + ":trans="
    string + ToStringObject(*state\trans)
    string + ")"
  Next
  
  ProcedureReturn string
EndProcedure


Procedure NFAAddState(*this.NFA, *value.Object, *isadd.Byte=#Null)
  Protected *state.NFAState, *result, isadd.Byte
  
  If (*isadd = #Null)
    *isadd = isadd
  EndIf
  
  *state = NewNFAState(*value)
  If (*this\state\size = 0)
    *this\start = *state
  EndIf
  
  *result = SetAddObject(*this\state, *state, *isadd)
  If (*isadd\b = #False)
;     FreeObject(*state\trans)
    FreeMemory(*state)
  Else
    *state\number = *this\state\size - 1
  EndIf
  
  ProcedureReturn *result
EndProcedure

; 
; Procedure NFAGetState(*this.NFA, *value.Object)
;   ProcedureReturn SetGetObject(*this\state, *value)
; EndProcedure
; 

Procedure pNFAElimateEpsTrans(*start.NFAState, *state.NFAState, eqclass.i, List *history.NFAState())
  Protected *trans.NFATrans
  Protected NewList *new.NFAState()
  Protected i
  
  ForEach *history()
    If (*history() = *state)
      ProcedureReturn 
    EndIf
  Next
  
  AddElement(*history())
  *history() = *state
  
  If (eqclass <> #Symbol_Epsilon)
    NFAStateAddTrans(*start, *state, eqclass)
  EndIf
  
  For i = 0 To *state\trans\size - 1
    *trans = *state\trans\object(i)
    
    If (*trans\value = #Symbol_Epsilon)
      If (*trans\goal\final > -1) And (eqclass = #Symbol_Epsilon)
        If (*trans\goal\final < *start\final) Or (*start\final = -1)
          *start\final = *trans\goal\final
        EndIf
      EndIf
      pNFAElimateEpsTrans(*start, *trans\goal, eqclass, *history())
      
    ElseIf (eqclass = #Symbol_Epsilon)
      ClearList(*new())
      pNFAElimateEpsTrans(*start, *trans\goal, *trans\value, *new())
    EndIf
    
  Next
  
  LastElement(*history())
  DeleteElement(*history())
EndProcedure

Procedure NFAEliminateEpsilons(*this.NFA)
  Protected *state.NFAState
  Protected *trans.NFATrans
  Protected NewList *history.NFAState()
  Protected i, k
  
  For i = 0 To *this\state\size - 1
    *state = *this\state\object(i)
    ClearList(*history())
    pNFAElimateEpsTrans(*state, *state, #Symbol_Epsilon, *history())
  Next
  
  For i = 0 To *this\state\size - 1
    *state = *this\state\object(i)
    
    For k = 0 To *state\trans\size - 1
      *trans = *state\trans\object(k)
      If (*trans\value = #Symbol_Epsilon)
        FreeObject(*trans)
        SetDelete(*state\trans, k)
        k - 1
      EndIf
    Next
  Next
EndProcedure


Procedure pNFAToDFAState(*this.NFA, *dfa.NFA, *curState.NFAState)
  Protected.Set *transSet, *stateSet, *dfaSet
  Protected.NFATrans *trans, *dfaTrans, *saveTrans
  Protected.NFAState *state, *dfaState, *saveState
  Protected isNew, isadd.Byte
  Protected NewList *recState.NFAState()
  Protected i, k
  
  *dfaSet = *curState\value
  *transSet = NewSet()
  
  For i = 0 To *dfaSet\size - 1
    *state = *dfaSet\object(i)
    *state = NFAAddState(*this, *state\value)
    
    If (*state\final > -1)
      If (*state\final < *curState\final) Or (*curState\final = -1)
        *curState\final = *state\final
      EndIf
    EndIf
    
    For k = 0 To *state\trans\size - 1
      *trans = *state\trans\object(k)
      *dfaTrans = NewNFATrans(NewNFAState(NewSet()), *trans\value)
      *dfaTrans\compare = @NFATransCompareSymbol()
      
      *saveTrans = SetAddObject(*transSet, *dfaTrans, isadd)
      If (isadd\b = #False)
        FreeObject(*dfaTrans)
      EndIf
      *dfaTrans = *saveTrans
      *dfaState = *dfaTrans\goal
      SetAddObject(*dfaState\value, CopyObject(*trans\goal))
    Next
  Next
  
  For i = 0 To *transSet\size - 1
    *dfaTrans = *transSet\object(i)
    *dfaState = *dfaTrans\goal
    *dfaState = NFAAddState(*dfa, *dfaState\value, isadd)
    
    NFAStateAddTrans(*curState, *dfaState, *dfaTrans\value)
    If (isadd\b <> #False)
      AddElement(*recState())
      *recState() = *dfaState
    EndIf
  Next
  FreeObject(*transSet)
  
  ForEach *recState()
    pNFAToDFAState(*this, *dfa, *recState())
  Next
EndProcedure


Procedure NFAToDFA(*this.NFA)
  Protected *dfa.NFA
  Protected *dfaSet.Set
  Protected *state.NFAState
  
  If (*this\state\size = 0)
    ProcedureReturn #Null
  EndIf
  
  *dfa = NewNFA(CopyObject(*this\charTable))
  *state = *this\start
  *dfaSet = NewSet()
  SetAddObject(*dfaSet, CopyObject(*state))
  
  pNFAToDFAState(*this, *dfa, NFAAddState(*dfa, *dfaSet))
  
  ProcedureReturn *dfa
EndProcedure


Procedure NewNFA(*charTable.CharTable)
  Protected *this.NFA
  
  *this = AllocateMemory(SizeOf(NFA))
  *this\free = @NFAFree()
  *this\copy = @StandardFreeFunction()
  *this\compare = @NFACompare()
  *this\tostring = @NFAToString()
  
  *this\state = NewSet()
  *this\charTable = *charTable
  
  ProcedureReturn *this
EndProcedure


Procedure.s pDFADotChr(i.i)
  Select i
  Case ' '
    ProcedureReturn "\\s"
  Case 9
    ProcedureReturn "\\t"
  Case 10,13
    ProcedureReturn "\\n"
  Case 'A' To 'Z', 'a' To 'z', '0' To '9'
    ProcedureReturn Chr(i)
  Case '"', 39
    ProcedureReturn "'\"+Chr(i)+"'"
  Case 32 To 127
    ProcedureReturn "'"+Chr(i)+"'"
  Case 256
    ProcedureReturn "<<EOF>>"
  Default
    ProcedureReturn "\\h"+RSet(Hex(i),2,"0")
  EndSelect
EndProcedure

Procedure DFAToDot(*dfa.NFA, file.s)
  Protected out, cntstate, i, s, str.s, istrans, islast
  Protected *state.NFAState, *trans.NFATrans
  Protected k, m
  
  out = CreateFile(#PB_Any, file)
  If (out <> 0)
    WriteStringFormat(out, #PB_UTF8)
    WriteStringN(out, "digraph G {")
    
    cntstate = *dfa\state\size
    For k = 0 To *dfa\state\size - 1
      *state = *dfa\state\object(k)
      
      WriteStringN(out, "")
      WriteString(out, "  S"+Str(*state\number)+" [shape=record, label="+Chr(34)+"{State "+Str(*state\number))
      If (*dfa\start = *state)
        WriteString(out, " - start")
      EndIf
      WriteString(out, " | ")
      WriteString(out, "Final = "+Str(*state\final))
      
      WriteStringN(out, "}"+Chr(34)+"]")
      
      Protected Dim char.i(cntstate, 256)
      For m = 0 To *state\trans\size - 1
        *trans = *state\trans\object(m)
        For i = 0 To 255
          If (*dfa\charTable\eqclass(i) = *trans\value)
            char(*trans\goal\number, i) = #True
          EndIf
        Next
        If (*trans\value = -1)
          char(*trans\goal\number, 256) = #True
        EndIf
      Next
      
      For s = 0 To cntstate-1
        str = ""
        istrans = #False
        islast = #False
        For i = 0 To 256
          If (char(s,i))
            If (islast = #False)
              If (str <> "")
                str + ", "
              EndIf
              str + pDFADotChr(i)
              islast = #True
            ElseIf (islast = #True)
              str + "-"
              islast = -1
            EndIf
            istrans = #True
          Else
            If (islast = -1)
              str + pDFADotChr(i-1)
            EndIf
            islast = #False
          EndIf
        Next
        If (islast = -1)
          str + pDFADotChr(i-1)
        EndIf
        If (istrans)
          WriteString(out, "  S"+Str(*state\number)+" -> S"+Str(s))
          WriteStringN(out, " [label="+Chr(34)+str+Chr(34)+"]")
        EndIf
      Next
    Next
    
    WriteStringN(out, "}")
    CloseFile(out)
  EndIf
EndProcedure

; 
; XIncludeFile "String.pb"
; Define *char.CharTable
; Define *nfa.NFA, *dfa.NFA, isadd.Byte
; Define.NFAState *q1, *q2, *q3
; 
; *char = NewCharTable()
; *char\eqclass('a') = 1
; *char\eqclass('b') = 2
; 
; *nfa = NewNFA(*char)
; *q1 = NFAAddState(*nfa, NewStdString("q1"), isadd)
; *q2 = NFAAddState(*nfa, NewStdString("q2"), isadd)
; *q3 = NFAAddState(*nfa, NewStdString("q3"), isadd)
; NFAStateAddTrans(*q1, *q2, 2, isadd)
; NFAStateAddTrans(*q1, *q3, #Symbol_Epsilon, isadd)
; NFAStateAddTrans(*q2, *q2, 1, isadd)
; NFAStateAddTrans(*q2, *q3, 1, isadd)
; NFAStateAddTrans(*q2, *q3, 2, isadd)
; NFAStateAddTrans(*q3, *q1, 1, isadd)
; *q1\final = #True
; 
; Debug "--------------"
; MyDebug(NFAToString(*nfa))
; NFAEliminateEpsilons(*nfa)
; 
; Debug "--------------"
; MyDebug(NFAToString(*nfa))
; 
; *dfa = NFAToDFA(*nfa)
; 
; Debug "------------"
; MyDebug(NFAToString(*dfa))
; 

; IDE Options = PureBasic 4.60 (Windows - x86)