
Declare.s RegExStateString(*this)

XIncludeFile "StringInclude.pb"
XIncludeFile "Error.pb"
XIncludeFile "CharTable.pb"
XIncludeFile "NFA.pb"
XIncludeFile "Stack.pb"


Enumeration -1 Step -1
  #RegEx_EndOfFile
  #RegEx_TypeStar
  #RegEx_TypeMinOne
  #RegEx_TypeOption
  #RegEx_TypeOr
  #RegEx_TypeAnd
  #RegEx_TypeEpsilon
  #RegEx_ParOpen  ;'('
  #RegEx_ParClose ;')'
EndEnumeration

#RegEx_FlagsNoCase = 1


Structure RegExState Extends Object
  regex.i
  state.i
EndStructure


Structure RegEx Extends Object
  StructureUnion
    eqclass.i
    type.i
  EndStructureUnion
  flags.i
  *charTable.CharTable
  List *item.RegEx()
  List *saveclass.RegEx()
EndStructure


;##############################################################################################################
;##### RegExState #####

Declare NewRegExState(regex.i, state.i)

Procedure RegExStateCopy(*this.RegExState)
  ProcedureReturn NewRegExState(*this\regex, *this\state)
EndProcedure


Procedure RegExStateCompare(*this.RegExState, *other.RegExState)
  Protected cmp
  
  cmp = *this\regex - *other\regex
  If (cmp <> 0)
    ProcedureReturn cmp
  EndIf
  
  ProcedureReturn *this\state - *other\state
EndProcedure


Procedure.s RegExStateToString(*this.RegExState)
  Protected string.s
  
  If (*this\regex = -1)
    If (*this\state = 0)
      string = "start"
    Else
      string = "end"
    EndIf
  Else
    string = "q"+Str(*this\regex)+"_"+Str(*this\state)
  EndIf
  
  ProcedureReturn string
EndProcedure


Procedure NewRegExState(regex.i, state.i)
  Protected *this.RegExState
  
  *this = AllocateMemory(SizeOf(RegExState))
  *this\free = @StandardFreeFunction()
  *this\copy = @RegExStateCopy()
  *this\compare = @RegExStateCompare()
  *this\tostring = @RegExStateToString()
  
  *this\regex = regex
  *this\state = state
  
  ProcedureReturn *this
EndProcedure


;##############################################################################################################
;##### RegEx #####

Procedure RegExCompare(*this.RegEx, *other.RegEx)
  Protected cmp
  
  cmp = *this\type - *other\type
  If (cmp <> 0)
    ProcedureReturn cmp
  EndIf
  
  cmp = ListSize(*this\item()) - ListSize(*other\item())
  If (cmp <> 0)
    ProcedureReturn cmp
  EndIf
  
  ResetList(*other\item())
  ForEach *this\item()
    NextElement(*other\item())
    cmp = *this\item()\compare(*this\item(), *other\item())
    If (cmp <> 0)
      ProcedureReturn cmp
    EndIf
  Next
  
  ProcedureReturn 0
EndProcedure


Declare NewRegEx(*charTable.CharTable=0, type.i=#RegEx_TypeOr)

Procedure RegExCopy(*this.RegEx)
  Protected *other.RegEx, i
  
  If (*this\charTable = #Null)
    *other = NewRegEx(#Null, *this\type)
  Else
    *other = NewRegEx(CopyObject(*this\charTable), *this\type)
  EndIf
  *other\flags = *this\flags
  
  ForEach *this\item()
    AddElement(*other\item())
    *other\item() = CopyObject(*this\item())
  Next
  
  ProcedureReturn *other
EndProcedure


Procedure RegExFree(*this.RegEx)
  ForEach *this\item()
    FreeObject(*this\item())
  Next
  ClearStructure(*this, RegEx)
  FreeMemory(*this)
EndProcedure


Procedure.s RegExToString(*this.RegEx)
  Protected string.s
  
  Select *this\type
  Case #RegEx_TypeAnd
    string = "AND"
  Case #RegEx_TypeOr
    string = "OR"
  Case #RegEx_TypeStar
    string = "STAR"
  Case #RegEx_TypeMinOne
    string = "MINONE"
  Case #RegEx_TypeOption
    string = "OPTION"
  Case #RegEx_TypeEpsilon
    string = "Epsilon"
  Default
    string + Str(*this\type)
  EndSelect
  
  If (ListSize(*this\item()) > 0)
    string + "["
  EndIf
  ForEach *this\item()
    If (ListIndex(*this\item()) > 0)
      string + ", "
    EndIf
    ;string + Chr(10)
    string + ToStringObject(*this\item())
  Next
  If (ListSize(*this\item()) > 0)
    ;string + Chr(10)
    string + "]"
  EndIf
  
  ProcedureReturn string
EndProcedure


Procedure pRegExItemToNFA(*regex.RegEx, *nfa.NFA, *start.NFAState, regidx, *stateidx.Integer)
  Protected.NFAState *endst, *savest
  
  Select *regex\type
  Case #RegEx_EndOfFile
    *endst = NFAAddState(*nfa, NewRegExState(regidx, *stateidx\i))
    *stateidx\i + 1
    NFAStateAddTrans(*start, *endst, #RegEx_EndOfFile)
    
  Case #RegEx_TypeAnd
    ForEach *regex\item()
      *start = pRegExItemToNFA(*regex\item(), *nfa, *start, regidx, *stateidx)
    Next
    *endst = *start
    
  Case #RegEx_TypeOr
    *endst = NFAAddState(*nfa, NewRegExState(regidx, *stateidx\i))
    *stateidx\i + 1
    ForEach *regex\item()
      If (*regex\item()\type >= 0)
        NFAStateAddTrans(*start, *endst, *regex\item()\eqclass)
      Else
        *savest = pRegExItemToNFA(*regex\item(), *nfa, *start, regidx, *stateidx)
        NFAStateAddTrans(*savest, *endst, #Symbol_Epsilon)
      EndIf
    Next
    
  Case #RegEx_TypeStar, #RegEx_TypeMinOne, #RegEx_TypeOption
    If (ListSize(*regex\item()) = 1)
      *savest = *start
      *start = NFAAddState(*nfa, NewRegExState(regidx, *stateidx\i))
      *stateidx\i + 1
      NFAStateAddTrans(*savest, *start, #Symbol_Epsilon)
      
      FirstElement(*regex\item())
      *endst = pRegExItemToNFA(*regex\item(), *nfa, *start, regidx, *stateidx)
      
      Select *regex\type
      Case #RegEx_TypeStar, #RegEx_TypeOption
        NFAStateAddTrans(*start, *endst, #Symbol_Epsilon)
      EndSelect
      Select *regex\type
      Case #RegEx_TypeStar, #RegEx_TypeMinOne
        NFAStateAddTrans(*endst, *start, #Symbol_Epsilon)
      EndSelect
    EndIf
    
  Case #RegEx_TypeEpsilon
    *endst = NFAAddState(*nfa, NewRegExState(regidx, *stateidx\i))
    *stateidx\i + 1
    NFAStateAddTrans(*start, *endst, #Symbol_Epsilon)
    
  Default
    *endst = NFAAddState(*nfa, NewRegExState(regidx, *stateidx\i))
    *stateidx\i + 1
    NFAStateAddTrans(*start, *endst, *regex\eqclass)
    
  EndSelect
  
  ProcedureReturn *endst
EndProcedure


Procedure RegExToNFA(*regex.RegEx)
  Protected *nfa.NFA, i
  Protected.NFAState *start, *endst
  Protected stateidx.Integer
  
  *nfa = NewNFA(*regex\charTable)
  *start = NFAAddState(*nfa, NewRegExState(-1, 0))
  
  If (*regex\type = #RegEx_TypeOr)
    i = 0
    ForEach *regex\item()
      stateidx\i = 0
      *endst = pRegExItemToNFA(*regex\item(), *nfa, *start, i, stateidx)
      *endst\final = i
      i + 1
    Next
  EndIf
  ProcedureReturn *nfa
EndProcedure


Procedure pRegExPriority(type.i)
  Select type
  Case #RegEx_TypeStar:   ProcedureReturn 9
  Case #RegEx_TypeMinOne: ProcedureReturn 9
  Case #RegEx_TypeOption: ProcedureReturn 9
  Case #RegEx_TypeAnd:    ProcedureReturn 8
  Case #RegEx_TypeOr:     ProcedureReturn 7
  Case #RegEx_ParOpen:    ProcedureReturn 0
  EndSelect
EndProcedure

Procedure pRegExCntOperand(type.i)
  Select type
  Case #RegEx_TypeStar: ProcedureReturn 1
  Case #RegEx_TypeMinOne: ProcedureReturn 1
  Case #RegEx_TypeOption: ProcedureReturn 1
  Default: ProcedureReturn 2
  EndSelect
EndProcedure

Procedure pRegExCharToType(char.c)
  Select char
  Case '|': ProcedureReturn #RegEx_TypeOr
  Case '*': ProcedureReturn #RegEx_TypeStar
  Case '+': ProcedureReturn #RegEx_TypeMinOne
  Case '?': ProcedureReturn #RegEx_TypeOption
  Case ')': ProcedureReturn #RegEx_ParClose
  EndSelect
EndProcedure

Procedure pRegExPushOper(*result.Stack, *item.RegEx)
  Protected count, *save.RegEx
  
  count = pRegExCntOperand(*item\type)
  While (count > 0)
    *save = StackPop(*result,#False)
    InsertElement(*item\item())
    *item\item() = *save
    count - 1
  Wend
  
  StackPush(*result, *item)
EndProcedure

Procedure pRegExEscapeGroup(char.c, Array charGr.i(1))
  Select char
  Case 'n'
    charGr(10) = #True
    charGr(13) = #True
  Case 't'
    charGr(9) = #True
  Default
    ProcedureReturn #False
  EndSelect
  ProcedureReturn #True
EndProcedure

Procedure RegExAddEval(*this.RegEx, string.s)
  Protected *char.Character, last.c
  Protected *regex.RegEx, *item.RegEx
  Protected *stack.Stack, *result.Stack
  Protected type.i, addeps.i, found.i, i.i, k.i, count.i
  Protected instring, inverse
  
  *stack = NewStack()
  *result = NewStack()
  addeps = #True
  
  *char = @string
  While ((*char\c <> 0) And ((*char\c <> '{') Or (instring = #True)))
    
    If (*char\c = '"')
      instring = 1 - instring
      *char + SizeOf(Character)
      Continue
    ElseIf (IsIgnoreChar(*char\c))
      *char + SizeOf(Character)
      Continue
    EndIf
    
    If (instring = #False)
      Select *char\c
      Case '|','*','?','+',')'
        type = pRegExCharToType(*char\c)
      Default ; character, '(', '[', '.' and "<<eof>>"
        type = #RegEx_TypeAnd
      EndSelect
    Else
      type = #RegEx_TypeAnd
    EndIf
    
    If (addeps)
      StackPush(*result, NewRegEx(#Null, #RegEx_TypeEpsilon))
    EndIf
    addeps = #False
    
    Select type
    Case #RegEx_TypeStar, #RegEx_TypeMinOne, #RegEx_TypeOption
      pRegExPushOper(*result, NewRegEx(#Null, type))
      
    Case #RegEx_ParClose
      found = #False
      While (StackIsEmpty(*stack) = #False)
        *item = StackPop(*stack,#False)
        If (*item\type = #RegEx_ParOpen)
          found = #True
          Break
        Else
          pRegExPushOper(*result, *item)
        EndIf
      Wend
      If (found = #False)
        StackFree(*stack)
        StackFree(*result)
        ErrorCode = #Error_NoParenthOpen
        ProcedureReturn 
      EndIf
      
    Default
      While (StackIsEmpty(*stack) = #False)
        *item = StackPeek(*stack)
        If (pRegExPriority(*item\type) < pRegExPriority(type))
          Break
        Else
          pRegExPushOper(*result, StackPop(*stack,#False))
        EndIf
      Wend
      StackPush(*stack, NewRegEx(#Null, type))
      If (*char\c = '(') And (instring = #False)
        StackPush(*stack, NewRegEx(#Null, #RegEx_ParOpen))
        type = #RegEx_ParOpen
      EndIf
      
    EndSelect
    
    Select type
    Case #RegEx_TypeAnd
      Protected Dim charGr.i(255)
      Protected Dim eqClass.i(255)
      
      If (instring = #False)
        Select *char\c
        Case '['
          *char + SizeOf(Character)
          inverse = #False
          If (*char\c = '^')
            inverse = #True
          EndIf
          *char - SizeOf(Character)
          last = 0
          Repeat
            *char + SizeOf(Character)
            Select *char\c
            Case 0, '{'
              ErrorCode = #Error_NoBracketClose
              StackFree(*stack)
              StackFree(*result)
              ProcedureReturn 
            Case '\'
              *char + SizeOf(Character)
              last = 0
              If pRegExEscapeGroup(*char\c, charGr())
                Continue
              EndIf
              
            Case '-'
              If (last > 0)
                *char + SizeOf(Character)
                If (*char\c <> ']')
                  If (last > *char\c)
                    ErrorCode = #Error_RangeOutOfOrdner
                    StackFree(*stack)
                    StackFree(*result)
                    ProcedureReturn 
                  EndIf
                  For i = last To *char\c
                    charGr(i) = #True
                  Next
                Else
                  *char - SizeOf(Character)
                EndIf
              EndIf
              
            Case ']'
              Break
            EndSelect
            
            charGr(*char\c) = #True
            last = *char\c
          ForEver
          
          If (inverse = #True)
            For i = 0 To 255
              charGr(i) = 1-charGr(i)
            Next
          EndIf
          
        Case ']'
          ErrorCode = #Error_NoBracketOpen
          StackFree(*stack)
          StackFree(*result)
          ProcedureReturn 
          
        Case '.'
          For i = 0 To 255
            charGr(i) = #True
          Next
          
        Case '<'
          If (LCase(PeekS(*char, 7)) = "<<eof>>")
            *item = NewRegEx(#Null, #RegEx_EndOfFile)
            StackPush(*result, *item)
            *char + SizeOf(Character)*7
            Continue
          Else
            ErrorCode = #Error_SyntaxError
            StackFree(*stack)
            StackFree(*result)
            ProcedureReturn 
          EndIf
          
        Case '\'
          *char + SizeOf(Character)
          If pRegExEscapeGroup(*char\c, charGr())
          Else
            charGr(*char\c) = #True
          EndIf
          
        Default
          charGr(*char\c) = #True
          
        EndSelect
      Else
        charGr(*char\c) = #True
      EndIf
      
      ; nocase
      If (*this\flags & #RegEx_FlagsNoCase)
        For i = 0 To 255
          If charGr(i)
            charGr(Asc(LCase(Chr(i)))) = #True
            charGr(Asc(UCase(Chr(i)))) = #True
          EndIf
        Next
      EndIf
      
      ; character without eqclass gets a new class
      *this\charTable\count + 1
      found = #False
      For i = 0 To 255
        If (charGr(i))
          If (*this\charTable\eqclass(i) = 0)
            *this\charTable\eqclass(i) = *this\charTable\count
            found = #True
          EndIf
          eqClass(*this\charTable\eqclass(i)) = #True
        EndIf
      Next
      If (found = #False)
        *this\charTable\count - 1
      EndIf
      
      count = *this\charTable\count
      For k = 1 To count
        If (eqClass(k))
          found = #False
          For i = 0 To 255
            If (charGr(i) = #False)
              If (*this\charTable\eqclass(i) = k)
                If (found = #False)
                  *this\charTable\count + 1
                EndIf
                found = #True
                *this\charTable\eqclass(i) = *this\charTable\count
              EndIf
            EndIf
          Next
          If (found = #True)
            ForEach *this\saveclass()
              If (*this\saveclass()\type = #RegEx_TypeOr)
                ForEach *this\saveclass()\item()
                  If (*this\saveclass()\item()\eqclass = k)
                    AddElement(*this\saveclass()\item())
                    *this\saveclass()\item() = NewRegEx(#Null, *this\charTable\count)
                  EndIf
                Next
              EndIf
            Next
          EndIf
        EndIf
      Next
      
      *item = NewRegEx(#Null, #RegEx_TypeOr)
      For k = 1 To *this\charTable\count
        If (eqClass(k))
          AddElement(*item\item())
          *item\item() = NewRegEx(#Null, k)
        EndIf
      Next
      StackPush(*result, *item)
      AddElement(*this\saveclass())
      *this\saveclass() = *item
      
    Case #RegEx_TypeStar, #RegEx_TypeMinOne, #RegEx_TypeOption, #RegEx_ParClose
    Default
      addeps = #True
    EndSelect
    
    *char + SizeOf(Character)
  Wend
  
  While (StackIsEmpty(*stack) = #False)
    *item = StackPop(*stack,#False)
    If (*item\type = #RegEx_ParOpen)
      ErrorCode = #Error_NoParenthClose
      StackFree(*stack)
      StackFree(*result)
      ProcedureReturn 
    EndIf
    pRegExPushOper(*result, *item)
  Wend
  
  AddElement(*this\item())
  *this\item() = StackPop(*result,#False)
  
  StackFree(*stack)
  StackFree(*result)
  
  ProcedureReturn (*char - @string) / SizeOf(Character) + 1
EndProcedure


Procedure pRegExMinimalize(*this.RegEx)
  Protected *item.RegEx
  
  ForEach *this\item(); (ab)c -> abc
    If (*this\item()\type = *this\type)
      *item = *this\item()
      DeleteElement(*this\item())
      ForEach *item\item()
        AddElement(*this\item())
        *this\item() = *item\item()
      Next
      ClearStructure(*item, RegEx)
      FreeMemory(*item)
      ProcedureReturn pRegExMinimalize(*this)
    EndIf
  Next
  
  Select *this\type
  Case #RegEx_TypeAnd; eaeb -> ab
    ForEach *this\item()
      If (*this\item()\type = #RegEx_TypeEpsilon) And (ListSize(*this\item()) > 1)
        FreeObject(*this\item())
        DeleteElement(*this\item())
      EndIf
    Next
  EndSelect
  
  Select *this\type
  Case #RegEx_TypeOr, #RegEx_TypeAnd; oper(a) -> a
    If (ListSize(*this\item()) = 1)
      FirstElement(*this\item())
      *item = *this
      *this = *this\item()
      ClearStructure(*item, RegEx)
      FreeMemory(*item)
      ProcedureReturn pRegExMinimalize(*this)
    EndIf
  EndSelect
  
  ForEach *this\item()
    *this\item() = pRegExMinimalize(*this\item())
  Next
  ProcedureReturn *this
EndProcedure

Procedure RegExMinimalize(*this.RegEx)
  If (*this\type = #RegEx_TypeOr)
    ForEach *this\item()
      *this\item() = pRegExMinimalize(*this\item())
    Next
  EndIf
EndProcedure


Procedure NewRegEx(*charTable.CharTable=0, type.i=#RegEx_TypeOr)
  Protected *this.RegEx
  
  *this = AllocateMemory(SizeOf(RegEx))
  InitializeStructure(*this, RegEx)
  *this\free = @RegExFree()
  *this\copy = @RegExCopy()
  *this\compare = @StandardCompareFunction()
  *this\tostring = @RegExToString()
  
  *this\eqclass   = type
  *this\charTable = *charTable
  
  ProcedureReturn *this
EndProcedure


; Define *char.CharTable, i
; Define *regex.RegEx, *nfa.NFA, *dfa.NFA
; 
; *regex = NewRegEx(NewCharTable())
; ; RegExAddEval(*regex, "includeeeeeeeeeeeeeeeeeeeeeeeeeeee")
; ; RegExAddEval(*regex, "class")
; ; RegExAddEval(*regex, "method")
; ; RegExAddEval(*regex, "null")
; RegExAddEval(*regex, "this")
; ; RegExAddEval(*regex, "super")
; RegExAddEval(*regex, "[0-9][0-9]*")
; RegExAddEval(*regex, "[a-zA-Z][a-zA-Z0-9_]*")
; 
; For i = 0 To #Symbol_Epsilon-1
;   If (*regex\charTable\eqclass(i))
;     Debug Chr(i)+": "+Str(*regex\charTable\eqclass(i))
;   EndIf
; Next
; Debug "---------"
; RegExMinimalize(*regex)
; MyDebug(RegExToString(*regex))
; Debug "---------"
; ; RegExString(*regex)
; ; End
; ;Debug "---------"
; 
; Define time, t1,t2,t3, i, *value
; time = ElapsedMilliseconds()
; *nfa = RegExToNFA(*regex)
; t1 = ElapsedMilliseconds()-time
; 
; time = ElapsedMilliseconds()
; NFAEliminateEpsilons(*nfa)
; MyDebug(NFAToString(*nfa))
; t2 = ElapsedMilliseconds()-time
; 
; time = ElapsedMilliseconds()
; *dfa = NFAToDFA(*nfa)
; t3 = ElapsedMilliseconds()-time
; MyDebug(NFAToString(*nfa))
; Debug ListSize(*dfa\state\object())
; 
; MessageRequester("Info", Str(t1)+","+Str(t2)+","+Str(t3))
; 

; IDE Options = PureBasic 4.60 (Windows - x86)