; Name: PBLexerGen (V.1.00)
; Author: Kevin Jasik (CSHW89)
; Date: 18.03.2012

IncludePath "Include\"
XIncludeFile "GetParameter.pb"
XIncludeFile "StringInclude.pb"
XIncludeFile "RegEx.pb"


#Version = 100


Global GraphvizProgram.s

Global Line_Nr
Global isOpenWin


Procedure LexerGenInfoWindow(dummy)
  OpenWindow(0, #PB_Ignore, #PB_Ignore, 404, 154, "LexerGen V."+StrF(#Version/100,2), #PB_Window_SystemMenu)
  EditorGadget(0, 2, 2, 400, 100, #PB_Editor_ReadOnly)
  
  isOpenWin = #True
  
  Repeat
  Until (WaitWindowEvent() = #PB_Event_CloseWindow)
  End
EndProcedure


Procedure LexerGenOpenFile(file.s)
  Protected in, out, pos, endpos
  Protected string.s, sttrim.s, output.s, key.s
  Protected *regex.RegEx, regidx, i, k, s, lexstate
  Protected *nfa.NFA, *dfa.NFA, *save, startstate
  Protected *state.NFAState, *trans.NFATrans
  Protected userDataText.s
  
  CreateThread(@LexerGenInfoWindow(), 0)
  Repeat: Until isOpenWin
  
  in = ReadFile(#PB_Any, file)
  If (in = 0)
    ProcedureReturn #False
  EndIf
  out = CreateFile(#PB_Any, GetTemporaryDirectory()+"lexergen.pb")
  If (out = 0)
    CloseFile(in)
    ProcedureReturn #False
  EndIf
  
  AddGadgetItem(0, -1, "Read inputfile...")
  
  *regex = NewRegEx(NewCharTable())
  regidx = 0
  Line_Nr = 0
  lexstate = 0
  userDataText = "*userData"
  
  output = "";"#__yylexer_generator = #True : "
  
  While (Eof(in) = #False)
    
    Line_Nr + 1
    string = ReadString(in)
    sttrim = MyTrim(string)
    
    If (Left(sttrim,2) = "%%")
      If (lexstate = 0)
        output + "XIncludeFile "+Chr(34)+"PBLexerGenInc.pbi"+Chr(34)+" : "
      EndIf
      lexstate + 1
      WriteStringN(out, output): output = ""
      Continue
    EndIf
    
    Select lexstate
    Case 0
      key = LCase(StartWith(sttrim))
      Select key
      Case "%nocase"
        *regex\flags | #RegEx_FlagsNoCase
        
      Case "%userdata"
        If (Mid(sttrim,Len(key)+1,1) <> " ")
          ErrorCode = #Error_SyntaxError
          ProcedureReturn #False
        EndIf
        userDataText = Mid(sttrim,Len(key)+2)
        
      Default
        output + string
      EndSelect
      WriteStringN(out, output): output = ""
      
    Case 1
      If (sttrim <> "")
        string = sttrim
        pos = RegExAddEval(*regex, string)
        If (ErrorCode <> 0)
          FreeObject(*regex)
          ProcedureReturn #False
        EndIf
        
        output + "Procedure __yylexerrun"+Str(regidx)+"("+userDataText+") : "
        
        string = Mid(string, pos+1)
        Repeat
          Repeat
            pos = FindStrMode(string, "", #FindMode_String)
            If (pos <> 0)
              If (UCase(Mid(string, pos+1, 5)) = "<EOF>")
                string = Left(string, pos-1) + "#yyEOF" + Mid(string, pos+6)
              Else
                string = Left(string, pos-1) + "#yyToken" + Mid(string, pos+1)
              EndIf
            EndIf
          Until (pos = 0)
          endpos = FindStrMode(string, "}", #FindMode_String)
          If (endpos = 0)
            If (Eof(in))
              ErrorCode = #Error_NoCurlyClose
              FreeObject(*regex)
              ProcedureReturn 0
            EndIf
            output + string
            WriteStringN(out, output): output = ""
            
            Line_Nr + 1
            string = ReadString(in)
          Else
            Break
          EndIf
        ForEver
        
        If (endpos = Len(string)) Or (Left(Trim(Mid(string,endpos+1)),1) = ";")
        Else
          ErrorCode = #Error_GarbageEndLine
          FreeObject(*regex)
          ProcedureReturn 0
        EndIf
        string = Left(string, endpos-1)
        output + string + " : ProcedureReturn -1 : EndProcedure"
        WriteStringN(out, output): output = ""
        
        regidx + 1
        
      Else
        WriteStringN(out, output): output = ""
      EndIf
      
    Case 2
      output + string
      WriteStringN(out, output): output = ""
      
    EndSelect
    
  Wend
  
  
  AddGadgetItem(0, -1, "Minimalize RegEx...")
  RegExMinimalize(*regex)
  
  AddGadgetItem(0, -1, "Transform RegEx to NFA...")
  *nfa = RegExToNFA(*regex)
  NFAEliminateEpsilons(*nfa)
  
  AddGadgetItem(0, -1, "Transform NFA to DFA...")
  *dfa = NFAToDFA(*nfa)
  
  AddGadgetItem(0, -1, "Write outputfile...")
  
  WriteStringN(out, "DataSection")
  WriteStringN(out, "  __yylexer_charclass:")
  output = ""
  For i = 0 To 255
    If (i % 16 = 0)
      output = "    Data.i "
    Else
      output + ","
    EndIf
    output + RSet(Str(*dfa\charTable\eqclass(i)),3," ")
    If (i % 16 = 15)
      ; If (i >= 32) And (i < 128)
      ;   output + "; "+Str(i-15)+"-"+Str(i)
      ; EndIf
      WriteStringN(out, output)
    EndIf
  Next
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  __yylexer_actionTable:")
  For i = 0 To *dfa\state\size - 1
    *state = *dfa\state\object(i)
    If (*dfa\start = *state)
      startstate = i
    EndIf
    
    Dim *transGoal.i(*dfa\charTable\count)
    
    For k = 0 To *state\trans\size - 1
      *trans = *state\trans\object(k)
      
      If (*trans\value = -1)
        *transGoal(0) = *trans\goal
      Else
        *transGoal(*trans\value) = *trans\goal
      EndIf
    Next
    
    *save = @*dfa\state\object()
    output = "    Data.i "
    For k = 0 To *dfa\charTable\count
      If (*transGoal(k) = #Null)
        output + RSet("-1", 4, " ")
      Else
        s = 0
        For s = 0 To *dfa\state\size - 1
          If (*dfa\state\object(s) = *transGoal(k))
            output + RSet(Str(s), 4, " ")
            Break
          EndIf
        Next
      EndIf
      output + ","
    Next
    If (*state\final = -1)
      output + " 0"
    Else
      output + " @__yylexerrun"+Str(*state\final)+"()"
    EndIf
    WriteStringN(out, output)
  Next
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  __yylexer_info:")
  WriteStringN(out, "    ; startstate, cntcharclass")
  WriteString (out, "    Data.i ")
  WriteString (out, Str(startstate)+", ")
  WriteStringN(out, Str(*dfa\charTable\count+1))
  WriteStringN(out, "EndDataSection")
  
  CloseFile(in)
  CloseFile(out)
  
  If (Param_PdfOutput <> "") And (Param_DotOutput = "")
    Param_DotOutput = GetTemporaryDirectory()+"lexergen.dot"
  EndIf
  If (Param_DotOutput <> "")
    AddGadgetItem(0, -1, "Write dotfile...")
    DFAToDot(*dfa, Param_DotOutput)
  EndIf
  If (Param_PdfOutput <> "")
    AddGadgetItem(0, -1, "Run Graphviz...")
    RunProgram(GraphvizProgram, "-Tpdf -o"+Param_PdfOutput+" "+Chr(34)+Param_DotOutput+Chr(34), "", #PB_Program_Wait)
    ; RunProgram(Param_PdfOutput)
  EndIf
  
  ProcedureReturn #True
EndProcedure


Procedure LexerGenStart()
  If OpenPreferences(GetPathPart(ProgramFilename())+"PBGenerators.ini")
    PreferenceGroup("Global")
    GraphvizProgram = ReadPreferenceString("graphviz", "")
    ClosePreferences()
  EndIf
  
  If (LexerGenOpenFile(Param_InputFile) = #False)
    AddGadgetItem(0, -1, "")
    AddGadgetItem(0, -1, "Error Line "+Str(Line_Nr)+":")
    AddGadgetItem(0, -1, GetErrorText())
    Repeat: Delay(10): ForEver
  EndIf
  AddGadgetItem(0, -1, "Done! No errors!")
  If (Param_OutputFile = "")
    Param_OutputFile = Param_InputFile + "_lexerout.pbl"
  EndIf
  CopyFile(GetPathPart(ProgramFilename())+"PBLexerGenInc.pbi", GetPathPart(Param_OutputFile)+"PBLexerGenInc.pbi")
  CopyFile(GetTemporaryDirectory()+"lexergen.pb", Param_OutputFile)
  DeleteFile(GetTemporaryDirectory()+"lexergen.pb")
EndProcedure


GetProgramParameter("LexerGen V."+StrF(#Version/100,2))
LexerGenStart()


; IDE Options = PureBasic 4.60 (Windows - x86)
; EnableXP
; Executable = PBLexerGen.exe
; CommandLine = "test.pbl"