; Name: PBParserGen (V.1.10)
; Author: Kevin Jasik (CSHW89)
; Date: 18.03.2012

IncludePath "Include\"
XIncludeFile "GetParameter.pb"
XIncludeFile "StringInclude.pb"
XIncludeFile "Grammar.pb"


#Version = 110


Global GraphvizProgram.s

Global Line_Nr
Global isOpenWin
Global WarningText.s


Procedure ParserGenInfoWindow(dummy)
  OpenWindow(0, #PB_Ignore, #PB_Ignore, 404, 204, "ParserGen V."+StrF(#Version/100,2), #PB_Window_SystemMenu)
  EditorGadget(0, 2, 2, 400, 200, #PB_Editor_ReadOnly)
  
  isOpenWin = #True
  
  Repeat
  Until (WaitWindowEvent() = #PB_Event_CloseWindow)
  End
EndProcedure


Procedure.s ParserGenReplaceValue(*rule.GrammarRule, string.s, *refnull.Integer)
  Protected pos, epos, char.c, index
  Protected replace.s, type.s
  
  *refnull\i = 0
  pos = -1
  Repeat
    pos = FindStrMode(string, "", #FindMode_String, pos+1)
    If (pos > 0)
      epos = pos+1
      char = Asc(Mid(string,epos,1))
      
      Select char
      Case ''
        *refnull\i = Line_Nr
        index = 0
        
      Case '1' To '9'
        index = 0
        epos - 1
        Repeat
          epos + 1
          index = index*10 + char-'0'
          char = Asc(Mid(string,epos+1,1))
        Until (char < '0') Or (char > '9')
        
      Default
        ErrorCode = #Error_RefInvalidChar
        ErrorText1 = Chr(char)
        ProcedureReturn ""
        
      EndSelect
      
      If (index > ListSize(*rule\item()))
        ErrorCode = #Error_RefIdxOutOfBounds
        ErrorText1 = ""+Str(index)
        ProcedureReturn ""
      EndIf
      
      replace = "__yyvalues("+Str(index)+")\"
      If (index = 0)
        type = *rule\start\type
      Else
        SelectElement(*rule\item(), index-1)
        type = *rule\item()\type
      EndIf
      
      If (type = "")
        ErrorCode = #Error_RefNoDeclaredType
        If (index = 0)
          ErrorText1 = ""
        Else
          ErrorText1 = ""+Str(index)
        EndIf
        ProcedureReturn ""
      EndIf
      
      replace + type
      string = Left(string, pos-1) + replace + Mid(string, epos+1)
      LastElement(*rule\item())
    EndIf
  Until (pos = 0)
  
  ProcedureReturn string
EndProcedure


Macro ParserAddGrammarRule(__rule__)
  If (lastisblock = #False)
    output + "Procedure __yyparserrun"+Str(runidx)+"("+userDataText+", Array __yyvalues.__yystructure(1)) : "
    output + " : ProcedureReturn 1 : EndProcedure : "
    runidx + 1
  EndIf
  __rule__\runidx = runidx-1
  GrammarAddRule(*grammar, __rule__)
  lastisblock = #False
EndMacro


Procedure ParserGenOpenFile(file.s)
  Protected in, out, debugout, i, k, count, pos, runidx, priority
  Protected string.s, sttrim.s, output.s, text.s, key.s, type.s
  Protected parserstate, instructure, tokenidx, unnameidx
  Protected *grammar.Grammar, *rule.GrammarRule, refnull.Integer
  Protected userDataText.s
  
  Protected *parser.NFA
  Protected cntstates, cnttermsym, cntnonterm
  
;   ParserGenInfoWindow(0)
  CreateThread(@ParserGenInfoWindow(), 0)
  Repeat: Until isOpenWin
  
  AddGadgetItem(0, -1, "Read inputfile...")
  
  in = ReadFile(#PB_Any, file)
  If (in = 0)
    ProcedureReturn #False
  EndIf
  out = CreateFile(#PB_Any, GetTemporaryDirectory()+"parsergen.pb")
  If (out = 0)
    CloseFile(in)
    ProcedureReturn #False
  EndIf
  
  *grammar = NewGrammar()
  GrammarAddTermSym(*grammar, "$EOF", "")
  GrammarAddNonTerm(*grammar, "$accept", "")
  
  GrammarAddNonTerm(*grammar, "error", "")
  *rule = #Null
  
  Line_Nr = 0
  parserstate = 0
  tokenidx = #GrammarItem_Term+1
  priority = 1
  instructure = #False
  runidx = 1
  userDataText = "*userData"
  
  output = ""
  
  While (Eof(in) = #False)
    
    Line_Nr + 1
    string = ReadString(in)
    sttrim = MyTrim(string)
    
    If (instructure)
      key = LCase(StartWith(sttrim))
      Select key
      Case "%endstructure"
        output + "EndStructure"
        instructure = #False
      Case "structureunion", "endstructureunion"
        output + string
      Case "%structureunion"
        output + "StructureUnion"
      Case "%endstructureunion"
        output + "EndStructureUnion"
      Case "array", "list", "map"
        ErrorCode = #Error_CantUsedCollect
        FreeObject(*grammar)
        ProcedureReturn #False
      Default
        output + string
      EndSelect
      WriteStringN(out, output): output = ""
      Continue
    EndIf
    
    If (Left(sttrim,2) = "%%")
      If (instructure)
        ErrorCode = #Error_WithoutEndStruc
        ProcedureReturn #False
      EndIf
      If (parserstate = 0)
        output + "XIncludeFile "+Chr(34)+"PBLexerGenInc.pbi"+Chr(34)+" : "
        output + "XIncludeFile "+Chr(34)+"PBParserGenInc.pbi"+Chr(34)+" : "
      EndIf
      parserstate + 1
      WriteStringN(out, output): output = ""
      Continue
    EndIf
    
    Select parserstate
    Case 0
      key = LCase(StartWith(sttrim))
      Select key
      Case "%token", "%type"
        pos = FindString(sttrim+" "," ",0)
        
        type = MyTrim(Mid(sttrim, Len(key)+1, pos-Len(key)-1))
        If (Left(type,1) = ".")
          type = Mid(type,2)
          If (type = "")
            ErrorCode = #Error_SyntaxError
            ProcedureReturn #False
          EndIf
        ElseIf (type <> "")
          ErrorCode = #Error_SyntaxError
          ProcedureReturn #False
        EndIf
        
        sttrim = Mid(sttrim, pos+1)
        
        count = CountString(sttrim, " ")
        For i = 0 To count
          text = StringField(sttrim, i+1, " ")
          If (text <> "")
            If (CheckName(text) = 0)
              ErrorCode = #Error_SyntaxError
              ProcedureReturn #False
            EndIf
            
            If (key = "%token")
              If FindMapElement(*grammar\termsym(), text)
                ErrorCode = #Error_DuplicateTermSym
                ErrorText1 = text
                ProcedureReturn #False
              EndIf
              GrammarAddTermSym(*grammar, text, type)
              output + "#yyToken"+text+" = "+Str(tokenidx)+" : "
              tokenidx + 1
            Else
              If FindMapElement(*grammar\nonterm(), text)
                ErrorCode = #Error_DuplicateNonTerm
                ErrorText1 = text
                ProcedureReturn #False
              EndIf
              GrammarAddNonTerm(*grammar, text, type)
            EndIf
            
          EndIf
        Next
        
      Case "%left", "%right"
        count = CountString(sttrim, " ")
        For i = 0 To count
          text = StringField(sttrim, i+1, " ")
          If (i = 0)
            If (text <> key)
              ErrorCode = #Error_SyntaxError
              ProcedureReturn #False
            EndIf
          Else
            If (text <> "")
              If (Left(text,1) = "'")
                If (Right(text,1) <> "'")
                  ErrorCode = #Error_SyntaxError
                  ProcedureReturn #False
                ElseIf (Len(text) <> 3)
                  ErrorCode = #Error_CharacterLength
                  ProcedureReturn #False
                EndIf
              Else
                If (CheckName(text) = 0)
                  ErrorCode = #Error_SyntaxError
                  ProcedureReturn #False
                EndIf
              EndIf
              If (FindMapElement(*grammar\termsym(), text) = #Null)
                GrammarAddTermSym(*grammar, text, "")
                If (Left(text,1) = "'")
                  *grammar\symbol(Asc(Mid(text,2,1))) = *grammar\termsym()\idx
                Else
                  output + "#yyToken"+text+" = "+Str(tokenidx)+" : "
                EndIf
                tokenidx + 1
              EndIf
              If (*grammar\termsym()\priority <> 0)
                ErrorCode = #Error_DuplicatePriority
                ErrorText1 = MapKey(*grammar\termsym())
                ProcedureReturn #False
              EndIf
              *grammar\termsym()\priority = priority
              If (key = "%left")
                *grammar\termsym()\assoc = #GrammarAssoc_Left
              Else
                *grammar\termsym()\assoc = #GrammarAssoc_Right
              EndIf
            EndIf
          EndIf
        Next
        priority + 1
        
      Case "%structure"
        instructure = #True
        output + "Structure __yystructure " + Mid(sttrim, Len(key)+1)
        
      Case "%endstructure"
        ErrorCode = #Error_WithoutStructure
        ProcedureReturn #False
        
      Case "%start"
        If (Mid(sttrim,Len(key)+1,1) <> " ")
          ErrorCode = #Error_SyntaxError
          ProcedureReturn #False
        EndIf
        text = Mid(sttrim,Len(key)+2)
        If (CheckName(text) = 0)
          ErrorCode = #Error_SyntaxError
          ProcedureReturn #False
        EndIf
        
        FindMapElement(*grammar\nonterm(), "$accept")
        *rule = NewGrammarRule(*grammar\nonterm())
        If (FindMapElement(*grammar\nonterm(), text) = #Null)
          GrammarAddNonTerm(*grammar, text, "")
        EndIf
        
        AddElement(*rule\item())
        *rule\item() = *grammar\nonterm()
        AddElement(*rule\item())
        FindMapElement(*grammar\termsym(), "$EOF")
        *rule\item() = *grammar\termsym()
        GrammarAddRule(*grammar, *rule)
        
      Case "%userdata"
        If (Mid(sttrim,Len(key)+1,1) <> " ")
          ErrorCode = #Error_SyntaxError
          ProcedureReturn #False
        EndIf
        userDataText = Mid(sttrim,Len(key)+2)
        
      Default
        output + string
      EndSelect
      WriteStringN(out, output): output = ""
      
    Case 1
      
      Protected *char.Character, *start.Character
      Protected *saverule.GrammarRule, symbol.s
      Protected instring, inblock, lastisblock
      
      If (*grammar\rules\size = 0)
        Line_Nr - 1
        ErrorCode = #Error_CantFindStart
        ProcedureReturn #False
      EndIf
      
      *rule = #Null
      Repeat
        If (sttrim <> "")
          *saverule = *rule
          If (inblock = #False)
            *rule = GrammarStartEval(*grammar, sttrim)
          Else
            *rule = 0
            *start = @sttrim
          EndIf
          If (*rule = 0)
            If (*saverule = 0)
              ErrorCode = 1000;//TODO
              ProcedureReturn #False
            EndIf
            *char = @sttrim
            *rule = *saverule
          Else
            If (*saverule <> 0)
              ParserAddGrammarRule(*saverule)
            EndIf
            *char = @sttrim + FindString(sttrim,":",0)*SizeOf(Character)
          EndIf
          instring = #False
          
          While (*char\c <> 0)
            
            If (*char\c = '"')
              If (inblock = #False)
                ErrorCode = #Error_SyntaxError
                ProcedureReturn #False
              EndIf
              instring = 1-instring
              If (instring = #True)
                *char + SizeOf(Character)
                Continue
              EndIf
            EndIf
            If (instring)
              *char + SizeOf(Character)
              Continue
            EndIf
            
            If (inblock = #False)
              If (IsNameChar(*char\c) Or (*char\c = 39))
                
                If (lastisblock = #True)
                  If (refnull\i <> 0)
                    Line_Nr = refnull\i
                    ErrorCode = #Error_RefNoDeclaredType
                    ErrorText1 = ""
                  EndIf
                  unnameidx + 1
                  GrammarAddNonTerm(*grammar, "$@"+Str(unnameidx), "")
                  AddElement(*rule\item())
                  *rule\item() = *grammar\nonterm()
                  *saverule = NewGrammarRule(*grammar\nonterm())
                  ParserAddGrammarRule(*saverule)
                  *saverule\cntvalues = ListSize(*rule\item())-1
                EndIf
                
                *start = *char
                *char + SizeOf(Character)
                If (*start\c = 39)
                  If (*char\c = 39)
                    ErrorCode = #Error_CharacterLength
                    ProcedureReturn #False
                  EndIf
                  *char + SizeOf(Character)
                  If (*char\c <> 39)
                    ErrorCode = #Error_CharacterLength
                    ProcedureReturn #False
                  EndIf
                  *char + SizeOf(Character)
                Else
                  While (IsNameChar(*char\c) Or IsNumberChar(*char\c))
                    *char + SizeOf(Character)
                  Wend
                EndIf
                symbol = PeekS(*start, (*char-*start) / SizeOf(Character))
                
                If (FindMapElement(*grammar\nonterm(), symbol))
                  AddElement(*rule\item())
                  *rule\item() = *grammar\nonterm()
                ElseIf (FindMapElement(*grammar\termsym(), symbol))
                  AddElement(*rule\item())
                  *rule\item() = *grammar\termsym()
                Else
                  If (*start\c = 39)
                    GrammarAddTermSym(*grammar, symbol, "")
                    AddElement(*rule\item())
                    *rule\item() = *grammar\termsym()
                    *grammar\symbol(Asc(Mid(symbol,2,1))) = *grammar\termsym()\idx
                  Else
                    ErrorCode = #Error_CantFindSymbol
                    ErrorText1 = symbol
                    ProcedureReturn #False
                  EndIf
                EndIf
                Continue
                
              ElseIf (IsIgnoreChar(*char\c))
              ElseIf (*char\c = '|')
                ParserAddGrammarRule(*rule)
                *rule = NewGrammarRule(*rule\start)
                
              ElseIf (*char\c = '{')
                inblock = #True
                *start = *char+SizeOf(Character)
                output + "Procedure __yyparserrun"+Str(runidx)+"("+userDataText+", Array __yyvalues.__yystructure(1)) : "
                runidx + 1
                
              ElseIf (*char\c = '%')
                symbol = ""
                *char + SizeOf(Character)
                If (LCase(PeekS(*char,4)) = "prec")
                  *char + SizeOf(Character)*4
                  If IsIgnoreChar(*char\c)
                    While IsIgnoreChar(*char\c)
                      *char + SizeOf(Character)
                    Wend
                    
                    If (IsNameChar(*char\c))
                      *start = *char
                      *char + SizeOf(Character)
                      While (IsNameChar(*char\c) Or IsNumberChar(*char\c))
                        *char + SizeOf(Character)
                      Wend
                      symbol = PeekS(*start, (*char-*start) / SizeOf(Character))
                      
                      If (FindMapElement(*grammar\termsym(), symbol))
                        *rule\precedence = *grammar\termsym()
                      Else
                        ErrorCode = #Error_CantFindSymbol
                        ErrorText1 = symbol
                        ProcedureReturn #False
                      EndIf
                    EndIf
                  EndIf
                EndIf
                If (symbol = "")
                  ErrorCode = #Error_SyntaxError
                  ProcedureReturn #False
                EndIf
                
              Else
                ErrorCode = #Error_SyntaxError
                ProcedureReturn #False
              EndIf
              
            Else
              If (*char\c <> '}') And (PeekC(*char+SizeOf(Character)) = 0)
                *char + SizeOf(Character)
              EndIf
              If (*char\c = '}') Or (*char\c = 0)
                If (*char\c = '}')
                  inblock = #False
                EndIf
                If (*char <> *start)
                  output + PeekS(*start, (*char-*start)/SizeOf(Character))
                  output = ParserGenReplaceValue(*rule, output, refnull)
                  If (ErrorCode <> 0)
                    ProcedureReturn #False
                  EndIf
                EndIf
                lastisblock = #True
                If (*char\c = 0)
                  Break
                EndIf
                output + " : ProcedureReturn 1 : EndProcedure : "
              EndIf
            EndIf
            
            *char + SizeOf(Character)
          Wend
        EndIf
       
        WriteStringN(out, output): output = ""
        If (Eof(in) = 0)
          Line_Nr + 1
          string = ReadString(in)
          sttrim = MyTrim(string)
        EndIf
        
        If (sttrim = "%%") Or (Eof(in))
          ParserAddGrammarRule(*rule)
          If (Eof(in))
            Break
          EndIf
          parserstate + 1
          WriteStringN(out, output): output = ""
          Break
        EndIf
      ForEver
      
    Case 2
      output + string
      WriteStringN(out, output): output = ""
      
    EndSelect
    
  Wend
  
  AddGadgetItem(0, -1, "Transform Grammar to LRParser...")
  *parser = GrammarToLRParser(*grammar)
  
  AddGadgetItem(0, -1, "Transform LRParser to SyntaxTable...")
  LRParserToSyntaxTable(*parser, *grammar)
  cntstates  = *parser\state\size
  cnttermsym = MapSize(*grammar\termsym())
  cntnonterm = MapSize(*grammar\nonterm())
  
  AddGadgetItem(0, -1, "Write outputfile...")
  
  WriteStringN(out, "DataSection")
  WriteStringN(out, "  __yyparser_actionTable:")
  
  For i = 0 To cntstates-1
    output = "    Data.i "
    For k = 0 To cnttermsym-1
      If (k > 0)
        output + ", "
      EndIf
      output + RSet("$"+Hex(parserState(i)\action(k)\idx),9," ")
    Next
    WriteStringN(out, output)
  Next
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  __yyparser_gotoTable:")
  
  For i = 0 To cntstates-1
    output = "    Data.i "
    For k = 0 To cntnonterm-1
      If (k > 0)
        output + ", "
      EndIf
      output + RSet("$"+Hex(parserState(i)\gotot(k)\idx),9," ")
    Next
    WriteStringN(out, output)
  Next
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  __yyparser_ruleTable:")
  For i = 0 To *grammar\rules\size - 1
    *rule = *grammar\rules\object(i)
    
    output = "    Data.i "
    output + Str(*rule\cntvalues)
    output + ", "
    If (*rule\cntvalues <> ListSize(*rule\item()))
      output + "1, "
    Else
      output + "0, "
    EndIf
    output + Str(#GrammarItem_NonTerm - *rule\start\idx)
    output + ", "
    If (i > 0)
      output + LSet("@__yyparserrun"+Str(i)+"()", 20, " ")
    Else
      output + LSet("0", 20, " ")
    EndIf
    output + "; "
    output + GrammarRuleTermString(*grammar, *rule\start\idx)+" ->"
    ForEach *rule\item()
      output + " " + GrammarRuleTermString(*grammar, *rule\item()\idx)
    Next
    WriteStringN(out, output)
  Next
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  __yyparser_charToToken:")
  output = ""
  For i = 0 To 255
    If (i % 16 = 0)
      output = "    Data.i "
    Else
      output + ","
    EndIf
    output + RSet(Str(*grammar\symbol(i)),4," ")
    If (i % 16 = 15)
      WriteStringN(out, output)
    EndIf
  Next
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  CompilerIf #yyDebug")
  WriteStringN(out, "    __yyparser_tokenText:")
  output = ""
  For i = 0 To cnttermsym-1
    If (i % 8 = 0)
      output = "      Data.s "
    Else
      output + ", "
    EndIf
    output + Chr(34)+GrammarRuleTermString(*grammar, #GrammarItem_Term + i)+Chr(34)
    If (i % 8 = 7)
      WriteStringN(out, output)
      output = ""
    EndIf
  Next
  If (output <> "")
    WriteStringN(out, output)
  EndIf
  WriteStringN(out, "  CompilerEndIf")
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  CompilerIf #yyDebug")
  WriteStringN(out, "    __yyparser_nonTermText:")
  output = ""
  For i = 0 To cntnonterm-1
    If (i % 8 = 0)
      output = "      Data.s "
    Else
      output + ", "
    EndIf
    output + Chr(34)+GrammarRuleTermString(*grammar, #GrammarItem_NonTerm - i)+Chr(34)
    If (i % 8 = 7)
      WriteStringN(out, output)
      output = ""
    EndIf
  Next
  If (output <> "")
    WriteStringN(out, output)
  EndIf
  WriteStringN(out, "  CompilerEndIf")
  
  WriteStringN(out, "  ")
  WriteStringN(out, "  __yyparser_info:")
  WriteStringN(out, "    ; startstate, errortoken, cntnonterm, cnttoken")
  WriteString (out, "    Data.i ")
  WriteString (out, Str(LRParserGetStart(*parser, *grammar))+", ")
  FindMapElement(*grammar\nonterm(), "error")
  WriteString (out, Str(#GrammarItem_NonTerm - *grammar\nonterm()\idx)+", ")
  WriteString (out, Str(cntnonterm)+", ")
  WriteStringN(out, Str(cnttermsym))
  WriteStringN(out, "EndDataSection")
  WriteStringN(out, "")
  
  CloseFile(in)
  CloseFile(out)
  
  If (Param_DebugOutput <> "")
    debugout = CreateFile(#PB_Any, GetTemporaryDirectory()+"parsergen_debug.txt")
  EndIf
  If (debugout <> 0)
    AddGadgetItem(0, -1, "Write debugfile...")
    LRParserDebugOutput(*parser, *grammar, debugout)
    CloseFile(debugout)
    CopyFile(GetTemporaryDirectory()+"parsergen_debug.txt", Param_DebugOutput)
  EndIf
  
  If (Param_PdfOutput <> "") And (Param_DotOutput = "")
    Param_DotOutput = GetTemporaryDirectory()+"parsergen.dot"
  EndIf
  If (Param_DotOutput <> "")
    AddGadgetItem(0, -1, "Write dotfile...")
    LRParserToDot(*parser, *grammar, Param_DotOutput)
  EndIf
  If (Param_PdfOutput <> "")
    AddGadgetItem(0, -1, "Run Graphviz...")
    RunProgram(GraphvizProgram, "-Tpdf -o"+Param_PdfOutput+" "+Chr(34)+Param_DotOutput+Chr(34), "", #PB_Program_Wait)
    ; RunProgram(Param_PdfOutput)
  EndIf
  
  WarningText = LRParserWarningText(*parser, *grammar)
  
  ProcedureReturn #True
EndProcedure


Procedure ParserGenStart()
  If OpenPreferences(GetPathPart(ProgramFilename())+"PBGenerators.ini")
    PreferenceGroup("Global")
    GraphvizProgram = ReadPreferenceString("graphviz", "")
    ClosePreferences()
  EndIf
  
  If (ParserGenOpenFile(Param_InputFile) = #False)
    AddGadgetItem(0, -1, "")
    AddGadgetItem(0, -1, "Error Line "+Str(Line_Nr)+":")
    AddGadgetItem(0, -1, GetErrorText())
    Repeat: Delay(10): ForEver
  EndIf
  
  AddGadgetItem(0, -1, "Done! No errors!")
  If (Param_OutputFile = "")
    Param_OutputFile = Param_InputFile + "_parserout.pby"
  EndIf
  CopyFile(GetPathPart(ProgramFilename())+"PBParserGenInc.pbi", GetPathPart(Param_OutputFile)+"PBParserGenInc.pbi")
  CopyFile(GetTemporaryDirectory()+"parsergen.pb", Param_OutputFile)
  DeleteFile(GetTemporaryDirectory()+"parsergen.pb")
  
  If (WarningText <> "")
    AddGadgetItem(0, -1, "")
    AddGadgetItem(0, -1, "Warning:")
    AddGadgetItem(0, -1, WarningText)
    Repeat: Delay(10): ForEver
  EndIf
EndProcedure


GetProgramParameter("ParserGen V."+StrF(#Version/100,2))
ParserGenStart()


; IDE Options = PureBasic 5.00 (Windows - x64)
; EnableXP
; Executable = PBParserGen.exe
; DisableDebugger
; CommandLine = "test\parser.pb" -o "test\parser_out.pb" -pdf "test\parser.pdf" -debug "test\debug.txt"